/****************************************************************************
 * drivers/wireless/generic/si4463.c
 *
 *   Copyright (C) 2017-2018 Sebastien Lorquet. All rights reserved.
 *   Author: Sebastien Lorquet <sebastien@lorquet.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name NuttX nor the names of its contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/

/****************************************************************************
 * Included Files
 ****************************************************************************/

#include <nuttx/config.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <nuttx/semaphore.h>

#include <nuttx/arch.h>
#include <nuttx/kmalloc.h>
#include <nuttx/wqueue.h>
#include <nuttx/semaphore.h>
#include <nuttx/fs/fs.h>
#include <nuttx/spi/spi.h>
#include <nuttx/wireless/generic/si4463.h>
#include <nuttx/wireless/generic/genradio.h>

#include "si4463.h"

/****************************************************************************
 * Pre-Processor Definitions
 ****************************************************************************/

#ifndef CONFIG_SCHED_HPWORK
#error High priority work queue required in this driver
#endif

#ifndef CONFIG_GENERICRADIO_SI4463_SPIMODE
#  define CONFIG_GENERICRADIO_SI4463_SPIMODE SPIDEV_MODE0
#endif

#ifndef CONFIG_GENERICRADIO_SI4463_FREQUENCY
#  define CONFIG_GENERICRADIO_SI4463_FREQUENCY 10000000 /* max 10 MHz */
#endif

/* Default to a 32-bit preamble length */
#ifndef CONFIG_GENERICRADIO_SI4463_DEFAULT_PREAMBLELEN
#define CONFIG_GENERICRADIO_SI4463_DEFAULT_PREAMBLELEN 32
#warning preamble length not defined, using 32
#elif CONFIG_GENERICRADIO_SI4463_DEFAULT_PREAMBLELEN == 0
#define CONFIG_GENERICRADIO_SI4463_DEFAULT_PREAMBLELEN 32
#warning preamble length was 0, using 32
#endif

/* Default to no preset base frequency */

#ifndef CONFIG_GENERICRADIO_SI4463_DEFAULT_FREQ
#define CONFIG_GENERICRADIO_SI4463_DEFAULT_FREQ 0
#endif

#ifndef CONFIG_GENERICRADIO_SI4463_DEFAULT_CHSPC
#define CONFIG_GENERICRADIO_SI4463_DEFAULT_CHSPC 0
#endif

#ifndef CONFIG_SPI_EXCHANGE
#  error CONFIG_SPI_EXCHANGE required for this driver
#endif

/* Values used to know what RF registers should be reprogrammed */

#define SI4463_APPLY_MODTYPE     (1 << 0)
#define SI4463_APPLY_BASEFREQ    (1 << 1)
#define SI4463_APPLY_DEVIATION   (1 << 2)
#define SI4463_APPLY_CHANSPACING (1 << 3)
#define SI4463_APPLY_DATARATE    (1 << 4)
#define SI4463_APPLY_BT          (1 << 5)
#define SI4463_APPLY_RXBW        (1 << 6)
#define SI4463_APPLY_TXFILT      (1 << 10)
#define SI4463_APPLY_RXFILT      (1 << 11)

/* Antenna switch states */

#define ANTENNA_IDLE 0
#define ANTENNA_TX   1
#define ANTENNA_RX   2

/****************************************************************************
 * Types
 ****************************************************************************/

/* A SI4463 device instance */

struct si4463_dev_s
{
  struct genradio_dev_s            genradio;    /* The public device instance */
  FAR struct spi_dev_s            *spi;         /* Underlying SPI bus */
  int                              spiid;       /* Device identifier within type SPIDEVTYPE_GENRADIO */
  struct work_s                    irqwork;     /* Interrupt continuation work queue support */
  FAR const struct si4463_lower_s *lower;       /* Low-level MCU-specific support */
  uint32_t                         xtal;        /* Frequency of the attached oscillator */
  bool                             hdr_done;    /* Set to TRUE when field 1 (header) has been received. */
#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
  uint8_t                          header[1];   /* Packet header, si4463 field 1, fixed size */
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2B ) || \
      defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2L )
  uint8_t                          header[2];
#endif
#if defined(CONFIG_GENERICRADIO_SI4463_ANTSWITCH_ALT)
  uint8_t                          papin;     /* Remember which pin activates the PA */
  uint8_t                          lnapin;    /* Remember which pin activates the LNA */
  bool                             activelow; /* TRUE if the PA and LNA use active low signals */
#endif

  /* Requested parameters, input to modem calculator */
  uint8_t                          requested_modtype; /* Requested modulation */
  uint32_t                         requested_datarate; /* Requested data rate in milli bps */
  uint32_t                         requested_basefreq; /* Requested base frequency in Hz */
  uint32_t                         requested_deviation; /* Requested FSK deviation in mHz */
  uint32_t                         requested_chanspacing; /* Requested channel spacing in mHz */
  uint32_t                         requested_bt; /* Requested GFSK parameter */
  uint32_t                         requested_rxbw; /* Requested RX bandwidth */
};

/****************************************************************************
 * Private function prototypes
 ****************************************************************************/

static int si4463_setbasefreq       (FAR struct genradio_dev_s *dev, 
                                     uint32_t Hz);
static int si4463_setchanspacing    (FAR struct genradio_dev_s *dev,
                                     uint32_t mHz);
static int si4463_setchannel        (FAR struct genradio_dev_s *dev,
                                     int32_t chan);
static int si4463_setdatarate       (FAR struct genradio_dev_s *dev,
                                     uint32_t bps);
static int si4463_ioctl             (FAR struct genradio_dev_s *dev, int cmd,
                                     unsigned long arg);
static int si4463_setmodulation     (FAR struct genradio_dev_s *dev,
                                     uint32_t mod);
static int si4463_setmodulationindex(FAR struct genradio_dev_s *dev,
                                     uint32_t ppm);
static int si4463_setdeviation      (FAR struct genradio_dev_s *dev,
                                     uint32_t mHz);
static int si4463_setfilterbt       (FAR struct genradio_dev_s *dev,
                                     uint32_t bt);
static int si4463_setpreamblelength (FAR struct genradio_dev_s *dev,
                                     uint32_t bits);
static int si4463_setsyncword       (FAR struct genradio_dev_s *dev,
                                     FAR uint8_t *data, uint32_t bits);
static int si4463_setrxbandwidth    (FAR struct genradio_dev_s *dev,
                                     uint32_t mHz);
static int si4463_getrssi           (FAR struct genradio_dev_s *dev,
                                     FAR int32_t *mB);
static int si4463_getlqi            (FAR struct genradio_dev_s *dev,
                                     FAR int32_t *lqi);
static int si4463_rxenable          (FAR struct genradio_dev_s *dev,
                                     int state);
static int si4463_settxpower        (FAR struct genradio_dev_s *dev,
                                     int32_t mB);
static void si4463_transmit         (FAR struct genradio_dev_s *dev,
                                     FAR const uint8_t *packet, size_t len);

/****************************************************************************
 * Global variables
 ****************************************************************************/

struct genradio_devops_s si4463_devops = 
{
  si4463_setbasefreq,        NULL,
  si4463_setchanspacing,     NULL,
  si4463_setchannel,         NULL,
  si4463_setdatarate,        NULL,
  si4463_ioctl,
  si4463_setmodulation,      NULL,
  si4463_setmodulationindex, NULL,
  si4463_setdeviation,       NULL,
  si4463_setfilterbt,        NULL,
  si4463_setpreamblelength,  NULL,
  si4463_setsyncword,        NULL,
  si4463_setrxbandwidth,     NULL,
  si4463_getrssi,
  si4463_getlqi,
  si4463_rxenable,
  si4463_settxpower,         NULL,
  si4463_transmit,
};

#ifdef CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG
#include "si4463_config.h"
static const uint8_t si4463_config_wds[] = RADIO_CONFIGURATION_DATA_ARRAY;
#endif

/****************************************************************************
 * Private functions
 ****************************************************************************/

/* htobel: convert uint32_t from host order to big endian
 * htobes: convert uint16_t from host order to big endian
 * betohl: convert uint32_t from big endian to host order
 * betohs: convert uint16_t from big endian to host order
 */

#ifdef CONFIG_ENDIAN_BIG
#define htobel(n) (n)
#define htobes(n) (n)
#define betohl(n) (n)
#define betohs(n) (n)
#else

static inline uint32_t htobel(uint32_t n)
{
  return ( n     &0xFF)<<24 |
         ((n>> 8)&0xFF)<<16 |
         ((n>>16)&0xFF)<< 8 |
         ((n>>24)&0xFF);
}

static inline uint16_t htobes(uint16_t n)
{
  return ( (n&0xFF)<<8 | n>>8 );
}

#define betohl(n) htobel(n)
#define betohs(n) htobes(n)
#endif

/* Hardware access routines */

/****************************************************************************
 * Name: si4463_lock
 *
 * Description:
 *   Acquire exclusive access to the shared SPI bus.
 *
 ****************************************************************************/

static void si4463_lock(FAR struct spi_dev_s *spi)
{
  SPI_LOCK        (spi, 1);
  SPI_SETBITS     (spi, 8);
  SPI_SETMODE     (spi, CONFIG_GENERICRADIO_SI4463_SPIMODE);
  SPI_SETFREQUENCY(spi, CONFIG_GENERICRADIO_SI4463_FREQUENCY);
}

/****************************************************************************
 * Name: si4463_unlock
 *
 * Description:
 *   Release exclusive access to the shared SPI bus.
 *
 ****************************************************************************/

static inline void si4463_unlock(FAR struct spi_dev_s *spi)
{
  SPI_LOCK(spi,0);
}

/****************************************************************************
 * Name: si4463_command
 *
 * Description:
 *   Send a command to a si4463 device
 *
 ****************************************************************************/

#define SI4463_RETRIES 2000

static int si4463_command(FAR struct si4463_dev_s *dev, uint8_t command,
                          FAR void *param   , uint16_t paramlen,
                          FAR void *response, uint16_t rsplen)
{
  uint8_t pollbuf[2];
  int retries = SI4463_RETRIES;
  int ret = 0;

  si4463_lock  (dev->spi);

  /* First, TX phase */

  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), true);

  /* Send command */
  SPI_SEND     (dev->spi, command);

  /* Send additional parameters */
  if(param && paramlen)
    {
      SPI_SNDBLOCK(dev->spi, (uint8_t*)param, paramlen);
    }

  /* Execute */
  SPI_SELECT(dev->spi, SPIDEV_GENRADIO(dev->spiid), false);

  /* Then, poll CTS until RX data is available */
  do
    {
      /* wait tSW = 80 ns */
      up_udelay(1);

      /* Store command at each attemt, it will get overwritten */
      pollbuf[0] = SI4463_CMD_READ_CMD_BUFF;
      pollbuf[1] = 0xFF;

      SPI_SELECT  (dev->spi, SPIDEV_GENRADIO(dev->spiid), true);
      SPI_EXCHANGE(dev->spi, pollbuf, pollbuf, 2);
      if (pollbuf[1] == 0xFF)
        {
          /* we're done */
          break;
        }
      else if(pollbuf[1] == 0x00)
        {
          /* command not completed yet */
          SPI_SELECT(dev->spi, SPIDEV_GENRADIO(dev->spiid), false);
          retries -= 1;
        }
      else
        {
          /* unexpected value! */
          _err("Unexpected rx value when reading CTS!\n");
          ret = -EIO; /* Transaction failed */
          goto unlock;
        }
    }
  while(retries != 0);

  if(!retries)
    {
      _err("Command timeout!\n");
      ret = -ETIMEDOUT;
      goto unlock;
    }

  /* Continue by reading the response data */
  if(response && rsplen)
    {
      SPI_RECVBLOCK(dev->spi, (uint8_t*)response, rsplen);
    }

unlock:
  SPI_SELECT(dev->spi, SPIDEV_GENRADIO(dev->spiid), false);
  si4463_unlock(dev->spi);

  return ret; /* Transaction finished */
}

/****************************************************************************
 * Name: si4463_getfrr
 *
 * Description:
 *   Return the value of the SI4463 fast registers. The transaction is executed
 *   as a single burst to avoid the DMA setup/teardown delays and be as fast
 *   as possible.
 *
 ****************************************************************************/

static int si4463_getfrr(FAR struct si4463_dev_s *dev, uint32_t firstreg,
                         FAR uint8_t *dest, size_t count)
{
  static const uint8_t cmds[4] =
    {
      SI4463_CMD_FRR_A_READ,
      SI4463_CMD_FRR_B_READ,
      SI4463_CMD_FRR_C_READ,
      SI4463_CMD_FRR_D_READ
    };
  uint8_t buf[5];

  if(firstreg > 3 || count < 0 || count > 4)
    {
      return -ERANGE;
    }

  buf[0] = cmds[firstreg];

  si4463_lock  (dev->spi);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), true);
  SPI_RECVBLOCK(dev->spi, buf, count + 1);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), false);
  si4463_unlock(dev->spi);
  memcpy(dest, buf + 1, count);

  return 0;
}

/****************************************************************************
 * Name: si4463_writetxfifo
 *
 * Description:
 *   Write data to the TX FIFO. There is no need to poll CTS.
 *
 ****************************************************************************/

static int si4463_writetxfifo(FAR struct si4463_dev_s *dev,
                              FAR const uint8_t *data, size_t len)
{
  if(len < 1)
    {
      return 0; /* Nothing to do and wommand will not work in that case */
    }

  si4463_lock  (dev->spi);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), true);
  SPI_SEND     (dev->spi, SI4463_CMD_WRITE_TX_FIFO);
  SPI_SNDBLOCK (dev->spi, data, len);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), false);
  si4463_unlock(dev->spi);
  return 0;
}

/****************************************************************************
 * Name: si4463_readrxfifo
 *
 * Description:
 *   Read data from the RX FIFO. There is no need to poll CTS.
 *
 ****************************************************************************/

static int si4463_readrxfifo(FAR struct si4463_dev_s *dev, FAR uint8_t *data,
                             size_t len)
{
  si4463_lock  (dev->spi);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), true);
  SPI_SEND     (dev->spi, SI4463_CMD_READ_RX_FIFO);
  SPI_RECVBLOCK(dev->spi, data, len);
  SPI_SELECT   (dev->spi, SPIDEV_GENRADIO(dev->spiid), false);
  si4463_unlock(dev->spi);
  return 0;
}

/****************************************************************************
 * Name: si4463_getchipstatus
 *
 * Description:
 *   Return information about the execution of the last command.
 *
 ****************************************************************************/

static int si4463_getchipstatus(FAR struct si4463_dev_s *dev)
{
  struct si4463_cmd_getchipstatus_s cmd;
  struct si4463_rsp_getchipstatus_s rsp;
  int ret;

  cmd.chip_clr_pend = ~CHIP_STATUS_CMD_ERROR; //clear cmd_error status only

  ret = si4463_command(dev, SI4463_CMD_GET_CHIP_STATUS,
                       &cmd, sizeof(struct si4463_cmd_getchipstatus_s),
                       &rsp, sizeof(struct si4463_rsp_getchipstatus_s));
  if (ret)
    {
      return ret;
    }

  if(rsp.chip_pend & CHIP_STATUS_CMD_ERROR)
    {
      _err("Command ID 0x%02X: error %d\n",
           rsp.cmd_err_cmd_id, rsp.cmd_err_status);
      if(!rsp.cmd_err_status)
        {
          return OK; /* Should not happen ??? */
        }
      switch(rsp.cmd_err_status)
        {
          case CMD_ERROR_BAD_COMMAND   /*16*/ : return -EBADMSG;
          case CMD_ERROR_BAD_ARG       /*17*/ : return -EINVAL;
          case CMD_ERROR_COMMAND_BUSY  /*18*/ : return -EBUSY;
          case CMD_ERROR_INVALID_STATE /*19*/ : return -EPROTO;
          case CMD_ERROR_BAD_BOOTMODE  /*49*/ : return -ENOSYS;
          case CMD_ERROR_BAD_PROPERTY  /*64*/ : return -ENOENT;
          case CMD_ERROR_TIMEOUT      /*240*/ : return -ETIMEDOUT;
          default                             : return -EIO;
        }
    }
  return 0;
}

/****************************************************************************
 * Name: si4463_getintstatus
 *
 * Description:
 *   Return information about the current pending interrupts, then clear them.
 *
 ****************************************************************************/

static int si4463_getintstatus(FAR struct si4463_dev_s *dev,
                               FAR uint8_t *phpend, FAR uint8_t *modempend,
                               FAR uint8_t *chippend)
{
  struct si4463_cmd_getintstatus_s cmd;
  struct si4463_rsp_getintstatus_s rsp;
  int ret;

  //Clear all pending ints after call
  cmd.ph_clr_pend    = 0;
  cmd.modem_clr_pend = 0;
  cmd.chip_clr_pend  = 0;

  ret = si4463_command(dev, SI4463_CMD_GET_INT_STATUS,
                       &cmd, sizeof(struct si4463_cmd_getintstatus_s),
                       &rsp, sizeof(struct si4463_rsp_getintstatus_s));
  if(ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if (ret != 0)
    {
      return ret;
    }

  if(phpend)
    {
      *phpend    = rsp.ph_pend;
    }

  if(modempend)
    {
      *modempend = rsp.modem_pend;
    }

  if(chippend)
    {
      *chippend  = rsp.chip_pend;
    }

  return 0;
}

/****************************************************************************
 * Name: si4463_getfifoinfo
 *
 * Description:
 *   Return the number of bytes in each fifo, and provide means to flush
 *   these FIFOs.
 *
 ****************************************************************************/

static int si4463_getfifoinfo(FAR struct si4463_dev_s *priv,
                              bool clean_tx, bool clean_rx,
                              FAR uint8_t *txlen, FAR uint8_t *rxlen)
{
  struct si4463_cmd_fifoinfo_s    cmd_fifo;
  struct si4463_rsp_fifoinfo_s    rsp_fifo;
  int ret;

  cmd_fifo.fifo = 0;
  if(clean_tx)
    {
      cmd_fifo.fifo |= FIFOINFO_CLEAR_TX;
    }
  if(clean_rx)
    {
      cmd_fifo.fifo |= FIFOINFO_CLEAR_RX;
    }

  ret = si4463_command(priv, SI4463_CMD_FIFO_INFO, &cmd_fifo, sizeof(cmd_fifo),
                       &rsp_fifo, sizeof(rsp_fifo));
  if(ret == 0)
    {
      ret = si4463_getchipstatus(priv);
    }
  if(ret != 0)
    {
      return ret;
    }

  if(txlen)
    {
      *txlen = rsp_fifo.tx_fifo_space;
    }
  if(rxlen)
    {
      *rxlen = rsp_fifo.rx_fifo_count;
    }

  return ret;
}

/****************************************************************************
 * Name: si4463_changestate
 *
 * Description:
 *   Transition the chip to a specific operating mode.
 *
 ****************************************************************************/

static int si4463_changestate(FAR struct si4463_dev_s *priv, uint8_t state)
{
  struct si4463_cmd_changestate_s cmd_chstate;
  int ret;

  cmd_chstate.next_state1 = state;

  ret = si4463_command(priv, SI4463_CMD_CHANGE_STATE,
        &cmd_chstate, sizeof(cmd_chstate), NULL, 0);

  if(ret == 0)
    {
      ret = si4463_getchipstatus(priv);
    }

  return ret;
}

/****************************************************************************
 * Name: si4463_setprop
 *
 * Description:
 *   Set the value of a SI4463 property (or successive properties)
 *
 ****************************************************************************/

static int si4463_setprop(FAR struct si4463_dev_s *dev, uint16_t prop,
                          FAR uint8_t* val, size_t len)
{
  struct si4463_cmd_setproperty_s cmd;
  int ret;

  if(len < 0 || len > 12)
    {
      return -E2BIG;
    }

  cmd.group      = (prop >> 8) & 0xFF;
  cmd.num_props  = len;
  cmd.start_prop = prop & 0xFF;
  memcpy(&cmd.data, val, len);

  ret = si4463_command(dev, SI4463_CMD_SET_PROPERTY,
                        &cmd,
                        sizeof(struct si4463_cmd_setproperty_s) - 
                            SI4463_SETPROP_DATALEN + len,
                        NULL, 0);
  if(ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }

  return ret;
}

/****************************************************************************
 * Name: si4463_setprop_byte
 *
 * Description:
 *   Set the value of a single byte SI4463 property
 *
 ****************************************************************************/

static inline int si4463_setprop_byte(FAR struct si4463_dev_s *dev,
                                      uint16_t prop, uint8_t val)
{
  return si4463_setprop(dev, prop, &val, 1);
}

/****************************************************************************
 * Name: si4463_setprop_short
 *
 * Description:
 *   Set the value of a short (big endian) SI4463 property
 *
 ****************************************************************************/

static inline int si4463_setprop_short(FAR struct si4463_dev_s *dev,
                                      uint16_t prop, uint16_t val)
{
  uint8_t arr[2];
  arr[0] = val >> 8;
  arr[1] = val & 0xFF;
  return si4463_setprop(dev, prop, arr, 2);
}

/****************************************************************************
 * Name: si4463_getprop
 *
 * Description:
 *   Return the value of a SI4463 property (or successive properties)
 *
 ****************************************************************************/

static int si4463_getprop(FAR struct si4463_dev_s *dev, uint16_t prop,
                          FAR uint8_t*val, size_t len)
{
  struct si4463_cmd_getproperty_s cmd;
  int ret;

  if(len < 0 || len > 16)
    {
      return -E2BIG;
    }

  cmd.group      = (prop >> 8) & 0xFF;
  cmd.num_props  = len;
  cmd.start_prop = prop & 0xFF;

  ret = si4463_command(dev, SI4463_CMD_GET_PROPERTY,
                        &cmd, sizeof(struct si4463_cmd_getproperty_s),
                        val, len);
  if(ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }

  return ret;
}

/****************************************************************************
 * Name: si4463_getadcreading
 *
 * Description:
 *   Return the values read on the internal or external ADC lines.
 *
 ****************************************************************************/

static int si4463_getadcreading(FAR struct si4463_dev_s *dev, int pin,
                                FAR uint16_t *gpio, FAR uint16_t *temp,
                                FAR uint16_t *vbat)
{
  struct si4463_cmd_getadcreading_s cmd;
  struct si4463_rsp_getadcreading_s rsp;
  int ret;

  cmd.adc_en  = 0;
  cmd.adc_cfg = 0;

  if (gpio)
    {
      if (pin < SI4463_IO0 || pin > SI4463_IO3)
        {
          _err("Invalid ADC pin %d\n", pin);
          return -EINVAL;
        }
      cmd.adc_en |=  ADC_EN_GPIO_EN;
      cmd.adc_en &= ~ADC_EN_GPIO_PIN_MASK;
      cmd.adc_en |=  ADC_EN_GPIO_PIN(pin);
    }

  if (temp)
    {
      cmd.adc_en |= ADC_EN_TEMPERATURE_EN;
    }

  if (vbat)
    {
      cmd.adc_en  |=  ADC_EN_BATTERY_VOLTAGE_EN;
      cmd.adc_cfg &= ~ADC_CFG_GPIOATT_MASK;
      cmd.adc_cfg |=  ADC_CFG_GPIOATT_3P2;
    }

  if (!cmd.adc_en)
    {
      _err("No ADC reading selected!\n");
      return -EINVAL;
    }

  /* Use the slow conversion rate */
  cmd.adc_cfg &= ~ADC_CFG_UDTIME_MASK;
  cmd.adc_cfg |=  ADC_CFG_UDTIME_305HZ;

  ret = si4463_command(dev, SI4463_CMD_GET_ADC_READING,
                        &cmd, sizeof(struct si4463_cmd_getadcreading_s),
                        &rsp, sizeof(struct si4463_rsp_getadcreading_s));
  if(ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }

  if (ret != 0)
    {
      return ret;
    }

  if(gpio)
    {
    *gpio = betohs(rsp.gpio_adc   ) & ADC_VALUE_MASK;
    }
  if(temp)
    {
    *temp = betohs(rsp.temp_adc   ) & ADC_VALUE_MASK;
    }
  if(vbat)
    {
    *vbat = betohs(rsp.battery_adc) & ADC_VALUE_MASK;
    }

  return 0;
}

/****************************************************************************
 * Name: si4463_packetconfig
 *
 * Description: Switch the TX/RX switch between IDLE/TX/RX positions. This is
 *   only used when we don't rely on the automatic switching made by the chip.
 *
 ****************************************************************************/
static int si4463_antennaswitch(FAR struct si4463_dev_s *dev, int pos)
{
#if defined(CONFIG_GENERICRADIO_SI4463_ANTSWITCH_ALT)
  int i;
  int ret;
  struct si4463_cmdrsp_gpiopincfg_s gpio;

  for(i=0;i<4;i++)
    {
      gpio.gpio[i]  = PINCFG_DONOTHING;
    }
  if(pos == ANTENNA_IDLE)
    {
      gpio.gpio[dev->papin ] = dev->activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0; // disabled
      gpio.gpio[dev->lnapin] = dev->activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0; // disabled
      _info("ant -> idle\n");
    }
  else if(pos == ANTENNA_TX)
    {
      gpio.gpio[dev->papin ] = dev->activelow ? PINCFG_DRIVE0 : PINCFG_DRIVE1; // enabled
      gpio.gpio[dev->lnapin] = dev->activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0; // disabled
      _info("ant -> tx\n");
    }
  else if(pos == ANTENNA_RX)
    {
      gpio.gpio[dev->papin ] = dev->activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0; // disabled
      gpio.gpio[dev->lnapin] = dev->activelow ? PINCFG_DRIVE0 : PINCFG_DRIVE1; // enabled
      _info("ant -> rx\n");
    }

  gpio.nirq         = PINCFG_DONOTHING;
  gpio.sdo          = PINCFG_DONOTHING;
  gpio.genconfig    = DRV_STRENGTH_HIGH;

  ret = si4463_command(dev, SI4463_CMD_GPIO_PIN_CFG,
                       &gpio, sizeof(struct si4463_cmdrsp_gpiopincfg_s),
                       &gpio, sizeof(struct si4463_cmdrsp_gpiopincfg_s));
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if(ret != 0)
    {
      _err("antenna switch setup failed (errno %d)\n", ret);
      return ret;
    }
#endif
  return 0;
}

/****************************************************************************
 * Name: si4463_packetconfig
 *
 * Description:
 *  Configure the packet handler. 3 configs are available at the moment.
 *   - 1 byte length in header, followed by data payload
 *   - 2 bytes big endian length in header, followed by data payload
 *   - 2 bytes lil endian length in header, followed by data payload
 * All packets end with a 2-byte CRC16.
 * The format is the same for TX and RX.
 * Address bytes in header are not managed yet.
 *
 ****************************************************************************/

static int si4463_packetconfig(FAR struct si4463_dev_s *dev)
{
  int ret;

  _info("Configuring packet handler\n");

  do
    {
      /* Configure CRC */

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_CRC_CONFIG,
                                PKT_CRC_CONFIG_SEED_FFFF | 
                                PKT_CRC_CONFIG_POLY_CCITT_16);
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_CONFIG1,
                                PKT_CONFIG1_CRC_INVERT);
      if(ret != 0)
        {
          break;
        }

      /* Configure field 1 as a one or two bytes length field */
#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
      ret = si4463_setprop_short(dev, SI4463_PROP_PKT_FIELD_1_LENGTH_0, 1);
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2B ) || \
      defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2L )
      ret = si4463_setprop_short(dev, SI4463_PROP_PKT_FIELD_1_LENGTH_0, 2);
#endif
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_FIELD_1_CONFIG, 0);
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_FIELD_1_CRC_CONFIG,
                                PKT_FIELD_CRC_CONFIG_START |
                                PKT_FIELD_CRC_CONFIG_ENABLE );
      if(ret != 0)
        {
          break;
        }

  /* Configure field 2 - length will be set before tx */
      ret = si4463_setprop_short(dev, SI4463_PROP_PKT_FIELD_2_LENGTH_0, 1);
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_FIELD_2_CONFIG, 0);
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_FIELD_2_CRC_CONFIG,
                                PKT_FIELD_CRC_CONFIG_ENABLE |
                                PKT_FIELD_CRC_CONFIG_CHECK |
                                PKT_FIELD_CRC_CONFIG_SEND);
      if(ret != 0)
        {
          break;
        }

  /* Disable fields 3-5 */
      ret = si4463_setprop_short(dev, SI4463_PROP_PKT_FIELD_3_LENGTH_0, 0);
      if(ret != 0)
        {
          break;
        }

  /* Configure rx length: in field 1, defines length of field 2. */

#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_LEN, 2 | PKT_LEN_IN_FIFO |
                                PKT_LEN_SIZE_1);
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2B )
      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_LEN, 2 | PKT_LEN_IN_FIFO |
                                PKT_LEN_SIZE_2 | PKT_LEN_ENDIAN_BIG);
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2L )
      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_LEN, 2 | PKT_LEN_IN_FIFO |
                                PKT_LEN_SIZE_2 | PKT_LEN_ENDIAN_LIL);
#endif
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_LEN_FIELD_SOURCE, 1);
      if(ret != 0)
        {
          break;
        }

      ret = si4463_setprop_byte(dev, SI4463_PROP_PKT_LEN_ADJUST, 0);
    }
  while(0);

  if(ret != 0)
    {
      _err("something failed\n", ret);
    }

  return ret;
}


/****************************************************************************
 * Name: si4463_initialize
 *
 * Description:
 *  Put the SI4463 in working order. Returns 0 if all is ok, else a negative
 *  errno value.
 *
 ****************************************************************************/

static int si4463_initialize(FAR struct si4463_dev_s *dev,
                             int papin, int lnapin, bool activelow)
{
  int ret;
#ifndef CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG
  uint8_t regs[2];
  struct si4463_cmd_pwrup_s pwrup;
#endif
  int i;
  struct si4463_cmdrsp_gpiopincfg_s gpio;
  struct si4463_rsp_partinfo_s part;
  struct si4463_rsp_funcinfo_s func;
  uint16_t vbat, temp;
  int32_t conv, iprt;

  /* check parameters */

  if(papin < SI4463_IO0 || papin > SI4463_IO3)
    {
      _err("Invalid PA control pin %d\n", papin);
      return -EINVAL;
    }

  if(lnapin < SI4463_IO0 || lnapin > SI4463_IO3)
    {
      _err("Invalid LNA control pin %d\n", lnapin);
      return -EINVAL;
    }

#ifndef CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG
  /* send POWER_UP command */

  _info("power up...\n");
  pwrup.boot_options = PWRUP_BOOTOPT_NO_PATCH | PWRUP_BOOTOPT_FUNC_PRO;
  pwrup.xtal_options = PWRUP_XTALOPT_XTAL;
  pwrup.xo_freq      = htobel(dev->xtal);
  ret = si4463_command(dev, SI4463_CMD_POWER_UP,
                       &pwrup, sizeof(struct si4463_cmd_pwrup_s), NULL, 0);
  if(ret != 0)
    {
      _err("Could not power up chip (errno %d)\n", ret);
      return ret;
    }

  /* Clean any pending interrupt */

  _info("clean up status...\n");
  ret = si4463_getintstatus(dev, NULL, NULL, NULL);
  if(ret != 0)
    {
      _err("Could not clean up status (errno %d)\n", ret);
      return ret;
    }
#else
  /* Use the WDS generated configuration settings */
  const uint8_t *config_ptr = si4463_config_wds;
  while(*config_ptr != 0)
    {
      uint8_t len = *config_ptr; config_ptr++;
      _info("Apply WDS config len= %d cmd= %02X\n", len, *config_ptr);
      ret = si4463_command(dev, *config_ptr, (uint8_t*)(config_ptr+1), len-1, NULL, 0);
      if (ret == 0)
        {
          ret = si4463_getchipstatus(dev);
        }
      if (ret != 0)
        {
          _err("Failed (ret %d)\n", ret);
          return ret;
        }
      config_ptr += len;
    }
#endif

  /* send PART_INFO command */

  _info("get part info...\n");
  ret = si4463_command(dev, SI4463_CMD_PART_INFO,
                       NULL, 0, &part, sizeof(struct si4463_rsp_partinfo_s));
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if (ret != 0)
    {
      _err("Could not get part info (errno %d)\n", ret);
      return ret;
    }

  part.part = betohs(part.part);
  part.id   = betohs(part.id  );
  _info("chiprev %02X part %04X pbuild %02X id %04X cust %02X romid %02X\n",
        part.chiprev, part.part, part.pbuild, part.id, part.customer,
        part.romid);

  /* send FUNC_INFO command */

  _info("get func info...\n");
  ret = si4463_command(dev, SI4463_CMD_FUNC_INFO,
                       NULL, 0, &func, sizeof(struct si4463_rsp_funcinfo_s));
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if (ret != 0)
    {
      _err("Could not get func info (errno %d)\n", ret);
      return ret;
    }

  func.patch = betohs(func.patch);
  _info("revext %02X revbranch %02X revint %02X patch %04X func %02X\n",
        func.revext, func.revbranch, func.revint, func.patch, func.func);

  if (part.part != 0x4463)
    {
      _err("Device not detected\n");
      return -EIO;
    }

#ifndef CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG
  /* Use a customized configuration */
  /* Global XO tune */
  regs[0] = 0x52; /* SI4463_PROP_GLOBAL_XO_TUNE (WDS value, 5.7 pF) */
  regs[1] = 0x00; /* SI4463_PROP_GLOBAL_CLK_CFG (WDS value, no clock outputs) */
  ret = si4463_setprop(dev, SI4463_PROP_GLOBAL_XO_TUNE, regs, 2);
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if (ret != 0)
    {
      _err("Could not define GLOBAL XO TUNE (errno %d)\n", ret);
      return ret;
    }

  /* Setup global config */

  _info("global config...\n");
  ret = si4463_setprop_byte(dev, SI4463_PROP_GLOBAL_CONFIG,
                           GLOBAL_CONFIG_RESERVED         |
                           GLOBAL_CONFIG_SEQUENCER_QUICK  |
                           GLOBAL_CONFIG_FIFO_HALFDUPLEX  |
                           GLOBAL_CONFIG_PROTOCOL_GENERIC |
                           GLOBAL_CONFIG_POWER_HIGH_PERF);
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if (ret != 0)
    {
      _err("Global config failed: %d\n", ret);
      return ret;
    }

  /* Setup common modem params */
  _info("TODO configure modem\n");

  /* Setup RSSI */
  _info("TODO configure RSSI\n");


#endif //CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG

  /* Configure PA/LNA pins and IRQ line */
  for(i=0;i<4;i++)
    {
      gpio.gpio[i]  = PINCFG_DONOTHING;
    }
#if defined(CONFIG_GENERICRADIO_SI4463_ANTSWITCH_ALT)
  dev->papin     = papin;
  dev->lnapin    = lnapin;
  dev->activelow = activelow;
  /* Just set both pins to idle */
  gpio.gpio[papin ] = activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0;
  gpio.gpio[lnapin] = activelow ? PINCFG_DRIVE1 : PINCFG_DRIVE0;
#else
  /* Use automatic PA/LNA activation by the chip - only active high */
  gpio.gpio[papin ] = activelow ? PINCFG_RX_STATE : PINCFG_TX_STATE;
  gpio.gpio[lnapin] = activelow ? PINCFG_TX_STATE : PINCFG_RX_STATE;
#endif
  gpio.nirq         = PINCFG_NIRQ_NIRQ;
  gpio.sdo          = PINCFG_DONOTHING;
  gpio.genconfig    = DRV_STRENGTH_HIGH;

  _info("configure gpios...\n");
  ret = si4463_command(dev, SI4463_CMD_GPIO_PIN_CFG,
                       &gpio, sizeof(struct si4463_cmdrsp_gpiopincfg_s),
                       &gpio, sizeof(struct si4463_cmdrsp_gpiopincfg_s));
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if(ret != 0)
    {
      _err("GPIO config failed (errno %d)\n", ret);
      return ret;
    }

  /* Read VBAT and TEMP */

  _info("Read ADCs...\n");
  ret = si4463_getadcreading(dev, 0, NULL, &temp, &vbat);
  if (ret == 0)
    {
      ret = si4463_getchipstatus(dev);
    }
  if(ret != 0)
    {
      _err(" ADC reading failed (errno %d)\n", ret);
      return ret;
    }
  conv = 3000L * (int32_t)vbat;
  conv /= 1280L;
  _info("Vbat: %ld mV [%02X]\n", conv, vbat);
  conv = (int32_t)temp * 899L*100L;
  conv /= 4096L;
  conv -= 293L * 100L;

  iprt = conv / 100;
  conv -= (iprt * 100);
  _info("Temp: %ld.%02ld deg C [%02X]\n", iprt, conv, temp);

  /* Setup packet handler */

  ret = si4463_packetconfig(dev);

  /* Setup fast read registers */

  /* FRR A is packet handler latched status, for quick int reaction */
  ret = si4463_setprop_byte(dev, SI4463_PROP_FRR_CTL_A_MODE, FRR_INT_PH_PEND);

  _info("Configure Interrupts\n");

  /* Setup IRQs: TX complete, data received, RX complete, RX error. */

  ret = si4463_setprop_byte(dev, SI4463_PROP_INT_CTL_ENABLE, INT_CTL_PH_INT | INT_CTL_MODEM_INT);
  if(ret != 0)
    {
      _err("SI4463_PROP_INT_CTL_ENABLE failed\n", ret);
      return ret;
    }

  ret = si4463_setprop_byte(dev, SI4463_PROP_INT_CTL_PH_ENABLE,
                                 INT_CTL_PH_RX_FIFO_ALMOST_FULL | /*0x01*/
//                               INT_CTL_PH_TX_FIFO_ALMOST_EMPTY| /*0x02*/
                                 INT_CTL_PH_CRC_ERROR           | /*0x08*/
                                 INT_CTL_PH_PACKET_RX           | /*0x10*/
                                 INT_CTL_PH_PACKET_SENT         | /*0x20*/
//                               INT_CTL_PH_FILTER_MISS         | /*0x40*/
//                               INT_CTL_PH_FILTER_MATCH        | /*0x80*/
                          0);
  if(ret != 0)
    {
      _err("SI4463_PROP_INT_CTL_PH_ENABLE failed\n", ret);
      return ret;
    }

  ret = si4463_setprop_byte(dev, SI4463_PROP_INT_CTL_MODEM_ENABLE,
                                 INT_CTL_MODEM_SYNC_DETECT      | /*0x01*/
                                 INT_CTL_MODEM_PREAMBLE_DETECT  | /*0x02*/
                                 INT_CTL_MODEM_INVALID_PREAMBLE | /*0x04*/
                                 INT_CTL_MODEM_RSSI             | /*0x08*/
                                 INT_CTL_MODEM_RSSI_JUMP        | /*0x10*/
                                 INT_CTL_MODEM_INVALID_SYNC     | /*0x20*/
                                 INT_CTL_MODEM_POSTAMBLE_DETECT | /*0x40*/
                                 INT_CTL_MODEM_RSSI_LATCH       | /*0x80*/
                                 0);

  if(ret != 0)
    {
      _err("SI4463_PROP_INT_CTL_MODEM_ENABLE failed\n", ret);
      return ret;
    }

  return ret; /* OK */
}

/****************************************************************************
 * Name: si4463_applysettings
 *
 * Description: Compute and apply registers
 *
 ****************************************************************************/
static int si4463_applysettings(FAR struct si4463_dev_s *priv, uint32_t what)
{
  int ret;
  uint64_t calc;
  uint8_t regset[7]; /*properties array*/
  uint32_t outdiv; // vco divider used at xl
  uint32_t ncomod;
  uint32_t txosr;

  if (!priv->requested_basefreq)
    {
      _err("Base frequency not yet set\n");
      return -EACCES;
    }

  /* Input params */
  _info("=============================\n");
  _info("Modem reconfiguration request\n");
  _info("Modulation type: %d\n", priv->requested_modtype);
  _info("Data rate      : %u mbps\n", priv->requested_datarate);
  _info("Base frequency : %u Hz\n", priv->requested_basefreq);
  _info("Deviation      : %u mHz\n", priv->requested_deviation);
  _info("Channel spacing: %u mHz\n", priv->requested_chanspacing);
  _info("RX bandwidth   : %u mHz\n", priv->requested_rxbw);
  _info("Gaussian BT    : %u.%u\n", priv->requested_bt>>16, priv->requested_bt&0xFFFF);
  _info("----------\n");
  /* ======================================================================== */
  /* MODULATION TYPE */
  /* ======================================================================== */


  if( (what & SI4463_APPLY_MODTYPE) == SI4463_APPLY_MODTYPE)
    {
      _info("Applying modulation settings\n");
      ret = si4463_getprop(priv, SI4463_PROP_SYNC_CONFIG, regset, 1);
      if (ret != 0)
        {
          _err("get SI4463_PROP_SYNC_CONFIG failed: %d\n", ret);
          return ret;
        }

      regset[0] &= ~SYNC_CONFIG_4FSK; /* Default to no 4GFSK in SYNC */
      regset[1]  = MOD_TYPE_MOD_SOURCE_PACKET; /* Data source = packet handler */

      switch(priv->requested_modtype)
        {
          case RADIO_MOD_CW:
            regset[1] = MOD_TYPE_MOD_TYPE_CW |      /* Data source is not PH ! */
                        MOD_TYPE_MOD_SOURCE_DIRECT;
            break;

          case RADIO_MOD_OOK:
          case RADIO_MOD_ASK:
            regset[1] |= MOD_TYPE_MOD_TYPE_OOK;
            _warn("OOK / ASK are not supported (yet)\n");
            break;

          case RADIO_MOD_2FSK:
            regset[1] |= MOD_TYPE_MOD_TYPE_2FSK;
            break;

          case RADIO_MOD_2GFSK:
            regset[1] |= MOD_TYPE_MOD_TYPE_2GFSK;
            break;

          case RADIO_MOD_4FSK:
            regset[1] |= MOD_TYPE_MOD_TYPE_4FSK;
            regset[0] |= SYNC_CONFIG_4FSK;
            break;

          case RADIO_MOD_4GFSK:
            regset[1] |= MOD_TYPE_MOD_TYPE_4GFSK;
            regset[0] |= SYNC_CONFIG_4FSK;
            break;

          default:
            _err("Unsupported modulation type %d\n", priv->requested_modtype);
            return -EINVAL;
        }

      ret = si4463_setprop_byte(priv, SI4463_PROP_SYNC_CONFIG, regset[0]);
      if (ret != 0)
        {
          _err("set SI4463_PROP_SYNC_CONFIG failed: %d\n", ret);
          return ret;
        }

      ret = si4463_setprop_byte(priv, SI4463_PROP_MODEM_MOD_TYPE, regset[1]);
      if (ret != 0)
        {
          _err("set SI4463_PROP_MODEM_MOD_TYPE failed: %d\n", ret);
          return ret;
        }
      priv->genradio.modulation = priv->requested_modtype;
    }

  /* ======================================================================== */
  /* DATA RATE */
  /* ======================================================================== */

  /*
   * Set the GFSK oversampling to 40 (<=25kbps), 20 (<200kbps) or 10 (same as WDS)
   * WDS sets the TXOSR to 10 for all FSK modes but we will do the same as GFSK.
   *
   * Set the NCO modulus to fxtal/10 (>200kbps) or fxtal (same as WDS)
   *
   * Reference: http://community.silabs.com/t5/tkb/articleprintpage/tkb-id/Proprietary%40tkb/article-id/110
   */

  /* Program module */
  if( (what & SI4463_APPLY_DATARATE) == SI4463_APPLY_DATARATE)
    {
      _info("Applying datarate settings\n");
      /* Define NCO_MODE according to datarate */
      ncomod = priv->xtal;
      if(priv->requested_datarate > 200000000)
        {
          /* Data rate > 200kbps, set fxtal/10 */
          ncomod /= 10;
        }

      /* Define the biggest possible oversample value */
      txosr = 40;
      regset[3] = MODEM_TX_NCO_MODE_TXOSR_40;
      if(priv->requested_datarate < 200000000)
        {
          txosr /= 2;
          regset[3] = MODEM_TX_NCO_MODE_TXOSR_20;
        }
      if(priv->requested_datarate <= 25000000)
        {
          txosr /= 2;
          regset[3] = MODEM_TX_NCO_MODE_TXOSR_10;
        }

      /* Compute rate value*/

      calc  = (uint64_t)priv->requested_datarate;
      calc *= (uint64_t)txosr;
      /* accum contains NCO_CLOCK_FREQ */
      calc *= (uint64_t)ncomod;
      calc /= (uint64_t)priv->xtal;
      calc /= 1000LLU; /* Because data rate in milli bps */

      _info("txosr %lu ncomod %lu ratediv %lu\n",txosr,ncomod,(uint32_t)calc);
      regset[0]  = (calc >> 16) & 0xFF;
      regset[1]  = (calc >>  8) & 0xFF;
      regset[2]  =  calc        & 0xFF;

      regset[3] |= (ncomod >> 24) & 0x03;
      regset[4]  = (ncomod >> 16) & 0xFF;
      regset[5]  = (ncomod >>  8) & 0xFF;
      regset[6]  =  ncomod        & 0xFF;

      ret = si4463_setprop(priv, SI4463_PROP_MODEM_DATA_RATE_0, regset, 7);
      if (ret != 0)
        {
          _err("set SI4463_PROP_MODEM_DATA_RATE_0 failed: %d\n", ret);
          return ret;
        }
      /* if MSK, also define freq deviation */

      if (priv->genradio.modulation == RADIO_MOD_MSK ||
          priv->genradio.modulation == RADIO_MOD_GMSK)
        {
          priv->requested_deviation = priv->requested_datarate >> 1; /* in mHz */
          what |= SI4463_APPLY_DEVIATION;
        }

      if(priv->requested_rxbw == 0)
        {
          _info("Rate: RW BW not specified, using auto settings\n");
          priv->requested_rxbw = priv->requested_datarate + 
                                 (priv->requested_deviation << 1);
          what |= SI4463_APPLY_RXFILT | SI4463_APPLY_TXFILT;
        }

      /* Recompute actual rate */
      calc *= 1000LLU;
      calc *= (uint64_t)priv->xtal;
      calc /= (uint64_t)ncomod;
      calc /= (uint64_t)txosr;

      priv->genradio.datarate = calc;
    }

  /* ======================================================================== */
  /* BASE FREQUENCY */
  /* ======================================================================== */

  /* Decide the VCO output divider according to the frequency band. The
  *  main VCO oscillates a 3.6 GHz giving the following output main freqs:
  *  4 -> 900 MHz
  *           --- half point 750 MHz
  *  6 -> 600 MHz
  *           --- half point 525 MHz
  *  8 -> 450 MHz
  *           --- half point 375 MHz
  * 12 -> 300 MHz
  *           --- half point 262 MHz
  * 16 -> 225 MHz
  *           --- half point 187 MHz
  * 24 -> 150 MHz
  *
  * However according to WDS the valid ranges are:
  * 142 - 175  (div = 24)
  * 284 - 350  (div = 12)
  * 420 - 525  (div =  8)
  * 850 - 1050 (div =  4)
  */
  if( (what & SI4463_APPLY_BASEFREQ) == SI4463_APPLY_BASEFREQ)
    {
      ret = si4463_getprop(priv, SI4463_PROP_MODEM_CLKGEN_BAND, &regset[4], 1);
      if (ret != 0)
        {
          return ret;
        }

      regset[4] |=  MODEM_CLKGEN_BAND_SY_SEL; /* Always high performance */
      regset[4] &= ~MODEM_CLKGEN_BAND_FVCO_DIV_MASK;
    }

  /* OUTDIV is computed inconditionnaly because required by other settings */

  if      (priv->requested_basefreq < 187500000)
    {
      outdiv     = 24;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_24;
    }
  else if (priv->requested_basefreq < 262500000)
    {
      outdiv     = 16;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_16;
    }
  else if (priv->requested_basefreq < 375000000)
    {
      outdiv     = 12;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_12;
    }
  else if (priv->requested_basefreq < 525000000)
    {
      outdiv     = 8;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_8;
    }
  else if (priv->requested_basefreq < 750000000)
    {
      outdiv     = 6;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_6;
    }
  else
    {
      outdiv     = 4;
      regset[4] |= MODEM_CLKGEN_BAND_FVCO_DIV_4;
    }

  _info("OUTDIV: %d\n", outdiv);

  if( (what & SI4463_APPLY_BASEFREQ) == SI4463_APPLY_BASEFREQ)
    {
      _info("Applying basefreq settings\n");

      /* fRF = (inte+ frac/2^19) (Npresc * xtal / outdiv)
       * (i<<19+f) = (fRF<<19) / (Npresc * xtal / outdiv)
       * (i<<19+f) = (outdiv * fRF<<19) / (Npresc * xtal8)
       */
      calc   = (uint64_t)priv->requested_basefreq;
      calc <<= 19;
      calc  *= outdiv;
      calc  /= (2LLU * (uint64_t)priv->xtal);

      if ((calc >> 19) > 0x80)
        {
          _err("cannot set frequency, incorrect integer divider\n");
          return -EINVAL;
        }

      /* we must remove 1 because the frac part has an implied 1 added. */
      regset[0] = ((calc >> 19) - 1) & 0x7F;
      regset[1] = (calc & 0x3FFFF) >> 16;
      regset[2] = (calc & 0x3FFFF) >>  8;
      regset[3] =  calc & 0x3FFFF       ;

#if 0
      _info("freq control : %u Hz, int %02x frac %02X %02X %02X\n",
            priv->requested_basefreq, 
            regset[0], regset[1], regset[2], regset[3]);
#endif

      ret = si4463_setprop_byte(priv, SI4463_PROP_MODEM_CLKGEN_BAND, regset[4]);
      if (ret != 0)
        {
          _err("set SI4463_PROP_MODEM_CLKGEN_BAND failed: %d\n", ret);
          return ret;
        }

      ret = si4463_setprop(priv, SI4463_PROP_FREQ_CONTROL_INTE, regset, 4);
      if (ret != 0)
        {
          _err("set SI4463_PROP_FREQ_CONTROL_INTE failed: %d\n", ret);
          return ret;
        }

      what |= SI4463_APPLY_RXFILT | SI4463_APPLY_TXFILT;

      /* Recompute actual frequency */

      calc  *= (2* (uint64_t)priv->xtal);
      calc  /= (uint64_t)outdiv;
      calc >>= 19;

      priv->genradio.basefreq = calc;
#warning check how to program SI4463_PROP_FREQ_CONTROL_VCOCNT_RX_ADJ, may change with IF
    }


  /* ======================================================================== */
  /* DEVIATION */
  /* ======================================================================== */


  /* Apply settings if requested */
  if( (what & SI4463_APPLY_DEVIATION) == SI4463_APPLY_DEVIATION)
    {
      _info("Applying deviation settings\n");

      calc  = (uint64_t)priv->requested_deviation << 19LLU;
      calc *= (uint64_t)outdiv;
      calc /= 2000LLU * (uint64_t)priv->xtal;

      if (calc >> 17)
        {
          _err("deviation too big\n");
          return -EINVAL;
        }

      /* truncate to 17 bits */
      calc &= 0x1FFFFLLU;
      _info("deviation divider: %lld\n", calc);
      regset[0] = (calc >> 16) & 0x01;
      regset[1] = (calc >>  8) & 0xFF;
      regset[2] =  calc        & 0xFF;

      ret = si4463_setprop(priv, SI4463_PROP_MODEM_FREQ_DEV_0, regset, 3);
      if (ret != 0)
        {
          _err("set SI4463_PROP_MODEM_FREQ_DEV_0 failed: %d\n", ret);
          return ret;
        }

      if(priv->requested_rxbw == 0)
        {
          _info("Deviation: RW BW not specified, using auto settings\n");
          priv->requested_rxbw = priv->requested_datarate +
                                 (priv->requested_deviation << 1);
          what |= SI4463_APPLY_RXFILT | SI4463_APPLY_TXFILT;
        }

      /* Recompute actual frequency deviation */

      calc *= 2000LLU * (uint64_t)priv->xtal;
      calc /= (uint64_t)outdiv << 19;
      priv->genradio.deviation = calc;
    }


  /* ======================================================================== */
  /* Channel spacing */
  /* ======================================================================== */

  if( (what & SI4463_APPLY_CHANSPACING) == SI4463_APPLY_CHANSPACING)
    {
      _info("Applying channel spacing settings\n");

      calc  = (uint64_t)outdiv << 19;
      calc *= (uint64_t)priv->requested_chanspacing;
      calc /= 2000LLU * (uint64_t)priv->xtal;

      if (calc > 0xFFFF)
        {
          _err("Cannot set this spacing, divider too big\n");
          return -EINVAL;
        }

      regset[0] = (calc>>8) & 0xFF;
      regset[1] =  calc     & 0xFF;

      ret = si4463_setprop(priv, SI4463_PROP_FREQ_CONTROL_CHANNEL_STEP_SIZE_0,
                           regset, 2);
      if (ret != 0)
        {
          _err("set SI4463_PROP_FREQ_CONTROL_CHANNEL_STEP_SIZE_0 failed: %d\n",
               ret);
          return ret;
        }

      /* Recompute actual chan spacing */

      calc *= 2000LLU * (uint64_t)priv->xtal;
      calc /= (uint64_t)outdiv << 19;

      priv->genradio.chanspacing = calc;
    }

  /* ======================================================================== */
  /* GFSK BT */
  /* ======================================================================== */

  /* ======================================================================== */
  /* Filters */
  /* ======================================================================== */
  if( (what & SI4463_APPLY_RXFILT) == SI4463_APPLY_RXFILT)
    {
      _info("Recompute RX filters TODO\n");
      /* Valid for CW and low rate xFSKs */
      static uint8_t synth_base[] = {0x2C, 0x0E, 0x0B, 0x04, 0x0C, 0x73, 0x03};

      /* Set by WDS for some high rates/deviation, TBD*/
      static uint8_t synth_fsk_500kbps_dev300khz[] = 
                     {0x01, 0x05, 0x0B, 0x05, 0x02, 0x00, 0x03};

      FAR uint8_t *synth_settings = synth_base;

      /* Apply synthetizer tuning values */

      ret = si4463_setprop(priv, SI4463_PROP_SYNTH_PFDCP_CPFF, synth_settings,
                           7);
      if (ret != 0)
        {
          _err("set SI4463_PROP_SYNTH_PFDCP_CPFF failed: %d\n", ret);
          return ret;
        }

    }

  if( (what & SI4463_APPLY_TXFILT) == SI4463_APPLY_TXFILT)
    {
      _info("Recompute TX filters TODO\n");
    }

  _info("----------\n");
  _info("Modem reconfiguration result\n");
  _info("Modulation type: %d\n"     , priv->genradio.modulation );
  _info("Data rate      : %u mbps\n", priv->genradio.datarate   );
  _info("Base frequency : %u Hz\n"  , priv->genradio.basefreq   );
  _info("Deviation      : %d mHz\n" , priv->genradio.deviation  );
  _info("Channel spacing: %d mHz\n" , priv->genradio.chanspacing);
  _info("RX bandwidth   : %d mHz\n" , priv->genradio.rxbandwidth);
  _info("-----------------------------\n");

  return 0;
}

/****************************************************************************
 * Driver functions
 ****************************************************************************/

/****************************************************************************
 * Name: si4463_setbasefreq
 *
 * Description: Define the main TX frequency, corresponding to channel 0.
 * 'fixed prescaler modulus in the PLL divider feedback path' is assumed to be
 * always set (high perf mode, Npresc = 2)
 *
 ****************************************************************************/
static int si4463_setbasefreq(FAR struct genradio_dev_s *dev, uint32_t Hz)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  priv->requested_basefreq = Hz;

  return si4463_applysettings(priv, SI4463_APPLY_BASEFREQ);
}

/****************************************************************************
 * Name: si4463_setchanspacing
 *
 * Description: Set the channel spacing
 * reg_spacing = (2^19 * outdiv * stepsize) / (Npresc * xtal)
 *
 ****************************************************************************/
static int si4463_setchanspacing(FAR struct genradio_dev_s *dev, uint32_t mHz)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  priv->requested_chanspacing = mHz;

  return si4463_applysettings(priv, SI4463_APPLY_CHANSPACING);
}

/****************************************************************************
 * Name: si4463_setchannel
 *
 * Description: Define the current radio channel. this info is just stored and
 *   then passed to the start_tx and start_rx radio commands as a single byte
 *
 ****************************************************************************/
static int si4463_setchannel(FAR struct genradio_dev_s *dev, int32_t chan)
{
  if (chan < 0 || chan > 255)
    {
      _err("incorrect requested channel %d\n", chan);
      return -EINVAL;
    }

  dev->channel = chan;
  return OK;
}

/****************************************************************************
 * Name: si4463_setdatarate
 *
 * Description: Define the module data rate.
 *
 ****************************************************************************/
static int si4463_setdatarate(FAR struct genradio_dev_s *dev, uint32_t mbps)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  priv->requested_datarate = mbps;

  return si4463_applysettings(priv, SI4463_APPLY_DATARATE);
}

/****************************************************************************
 * Name: si4463_setmodulation
 *
 * Description:
 *
 ****************************************************************************/
static int si4463_setmodulation(FAR struct genradio_dev_s *dev, uint32_t mod)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  /* Define the sync word modulation */

  priv->requested_modtype = mod;

  return si4463_applysettings(priv, SI4463_APPLY_MODTYPE);
}

/****************************************************************************
 * Name: si4463_setmodulationindex
 *
 * Description: Set the modulation index.
 *   For ASK: amplitude modulation index (not supported yet).
 *   For FSK: h = 2 * deviation[from carrier] / symbolrate
 *
 ****************************************************************************/
static int si4463_setmodulationindex(FAR struct genradio_dev_s *dev,
                                     uint32_t ppm)
{
  if(dev->modulation < RADIO_MOD_FSK_LAST)
    {
    /* h = 2 * dev / rate
     * 2 * dev = h * rate
     * dev = h * rate / 2
     * modindex in ppm
     * datarate  in bps/1000
     * result deviation in mHz
     */
    uint32_t calculated = (uint32_t)( ((uint64_t)ppm * 
                                       (uint64_t)dev->datarate) / 2000000LLU );

    _info("Request FSK modindex %u ppm -> dev %u mHz\n", ppm, calculated);

    return si4463_setdeviation(dev, calculated);
    }

  return -ENOTTY;
}

/****************************************************************************
 * Name: si4463_setdeviation
 *
 * Description: setup the frequency deviation from the carrier.
 * For MSK this is forbidden, the call is made while setting the data rate.
 * For 4FSK this is the external deviation, internal one is d/3
 *
 ****************************************************************************/
static int si4463_setdeviation(FAR struct genradio_dev_s *dev, uint32_t mHz)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  if(dev->modulation < RADIO_MOD_FSK_LAST)
    {
      priv->requested_deviation = mHz;  

      return si4463_applysettings(priv, SI4463_APPLY_DEVIATION);
    }

  /* other modulations are not allowed to use a deviation */
  return -EINVAL;
}

/****************************************************************************
 * Name: si4463_setfilterbt
 *
 * Description:
 *
 ****************************************************************************/
static int si4463_setfilterbt(FAR struct genradio_dev_s *dev, uint32_t bt)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  priv->requested_bt = bt;

  return si4463_applysettings(priv, SI4463_APPLY_BT);
}

/****************************************************************************
 * Name: si4463_setpreamblelength
 *
 * Description:
 *   Set the number of preamble bits, granularity is 4 bits.
 *   The SI4463 has numerous options
 *   to control the preamble, we are only using a standard preamble pattern.
 *   In RX mode, N-2 consecutive bits must be detected to detect
 *   a valid preamble.
 *   Channel hopping is not used in this driver, so there is no need for a
 *   preamble timeout. Postamble is not enabled.
 *
 ****************************************************************************/
static int si4463_setpreamblelength(FAR struct genradio_dev_s *dev,
                                    uint32_t bits)
{
  int ret;
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;
  uint8_t nibbles = bits >> 2;
  
  if ( bits != (nibbles<<2))
    {
      _err("Unsupported preamble length %d, must be a multiple of 4 bits\n",
           bits);
      return -EINVAL;
    }

  ret = si4463_setprop_byte(priv, SI4463_PROP_PREAMBLE_CONFIG,
                            PREAMBLE_CONFIG_RX_PREAM_SRC_STANDARD |
                            PREAMBLE_CONFIG_PREAM_FIRST_0         |
                            PREAMBLE_CONFIG_LENGTH_CONFIG_NIBBLE  |
                            PREAMBLE_CONFIG_MAN_DISABLE           |
                            PREAMBLE_CONFIG_PRE_0101
                            );
  if (ret != 0)
    {
      _err("set SI4463_PROP_PREAMBLE_CONFIG failed: %d\n", ret);
      return ret;
    }
    
  ret = si4463_setprop_byte(priv, SI4463_PROP_PREAMBLE_CONFIG_STD_1,
                            PREAMBLE_CONFIG_STD_1_SKIP_SYNC_TIMEOUT_DISABLE |
                            PREAMBLE_CONFIG_STD_1_RX_THRES(bits-2)
                            );
  if (ret != 0)
    {
      _err("set SI4463_PROP_PREAMBLE_CONFIG_STD_1 failed: %d\n", ret);
      return ret;
    }
                            
  ret = si4463_setprop_byte(priv, SI4463_PROP_PREAMBLE_TX_LENGTH, nibbles);
  if (ret != 0)
    {
      _err("SI4463_PROP_PREAMBLE_TX_LENGTH failed");
      return ret;
    }
  
  dev->preamblelen = bits;
  
  return 0;
}

/****************************************************************************
 * Name: si4463_setsyncword
 *
 * Description:
 *   Define the sync word transmitted between the preamble and the packet.
 *   It is used on the receive side to detect the start of a packet.
 *   Lengths 2-32 in multiple of 2 bits are supported.
 *   Manchester is never used, 4fsk is retained if the modulation is defined to
 *   4fsk, but is not modified in this function.
 *
 ****************************************************************************/
static int si4463_setsyncword(FAR struct genradio_dev_s *dev, FAR uint8_t *data,
                              uint32_t bits)
{
  int ret;
  uint8_t prop;
  uint8_t subbits,bytes;
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  /* Preliminary checks */
  if(bits < 2 || bits > 32 || (bits & 1) )
    {
      return -EINVAL;
    }

  /* Compute number of full bytes of sync bits, minus one
   *  2- 8 : 0
   * 10-16 : 1
   * 18-24 : 2
   * 26-32 : 3 */
  bytes = (bits-1)>>3;

  /* Compute the number of bits to be subtracted from the byte length */
  subbits = (8 - (bits & 7)) >> 1;
  if(subbits==4)
    {
      subbits = 0;
    }

  _info("sync len %d -> bytes=%d, subbits=%d\n", bits, bytes, subbits);

  /* First, get the config and save the 4FSK bit */
  ret = si4463_getprop(priv, SI4463_PROP_SYNC_CONFIG, &prop, 1);
  if (ret != 0)
    {
      _err("get SI4463_PROP_SYNC_CONFIG failed: %d\n", ret);
      return ret;
    }
  prop &= SYNC_CONFIG_4FSK;

  /* Modify required config bits */

  ret = si4463_setprop_byte(priv, SI4463_PROP_SYNC_CONFIG,
                            SYNC_CONFIG_LENGTH(bytes) |
                            SYNC_CONFIG_RX_ERRORS(2) |
                            prop);
  if (ret != 0)
    {
      _err("set SI4463_PROP_SYNC_CONFIG failed: %d\n", ret);
      return ret;
    }

  ret = si4463_setprop_byte(priv, SI4463_PROP_SYNC_CONFIG2,
                            SYNC_CONFIG2_LENGTH_SUB(subbits));
  if (ret != 0)
    {
      _err("set SI4463_PROP_SYNC_CONFIG2 failed: %d\n", ret);
      return ret;
    }

  /* Set the actual bits */
  ret = si4463_setprop(priv, SI4463_PROP_SYNC_BITS_0, data, bytes+1);
  if (ret != 0)
    {
      _err("set SI4463_PROP_SYNC_BITS_0 failed: %d\n", ret);
      return ret;
    }

  dev->syncwordlen = bits;
  memcpy(dev->syncword, data, BYTESFORBITS(bits));

  return 0;
}

/****************************************************************************
 * Name: si4463_setrxbandwidth
 *
 * Description: Define a manual RX bandwidth. Value zero means the required
 *   bandwidth is automatically calculated from other parameters.
 *
 ****************************************************************************/
static int si4463_setrxbandwidth(FAR struct genradio_dev_s *dev, uint32_t mHz)
{
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  priv->requested_rxbw = mHz;

  return si4463_applysettings(priv, SI4463_APPLY_RXBW);
}

/****************************************************************************
 * Name: si4463_getrssi
 *
 * Description:
 *
 ****************************************************************************/
static int si4463_getrssi(FAR struct genradio_dev_s *dev, FAR int32_t *mB)
{
  return -ENOTTY;
}

/****************************************************************************
 * Name: si4463_getlqi
 *
 * Description:
 *
 ****************************************************************************/
static int si4463_getlqi(FAR struct genradio_dev_s *dev, FAR int32_t *lqi)
{
  return -ENOTTY;
}

/****************************************************************************
 * Name: si4463_settxpower
 *
 * Description: Adjust the PA and TX power. There is no simple relation
 * between the output power and the PA setting, it MUST be calibrated.
 *
 ****************************************************************************/
static int si4463_settxpower(FAR struct genradio_dev_s *dev, int32_t mB)
{
  uint8_t regs[4];
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;
  
  regs[0] = 0x08;
  regs[1] = mB;
  regs[2] = 0x00;
  regs[3] = 0x3D;

  _info("txpw setting: %02X\n", regs[1]);

  return si4463_setprop(priv, SI4463_PROP_PA_MODE, regs, 4);
}

/****************************************************************************
 * Name: si4463_rxenable
 *
 * Description: Enable or disable the interrupt related to packet reception.
 * If the interrupt is enabled, also go to the rx state so the next valid
 * packet on air will be received.
 *
 ****************************************************************************/
static int si4463_rxenable(FAR struct genradio_dev_s *dev, int state)
{
  int ret;
  struct si4463_cmd_startrx_s cmd_rx;
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  if(!state)
    {
      _info("disable rx\n");
      /* We are called to STOP receiving */
      ret = si4463_antennaswitch(priv, ANTENNA_IDLE);
      if (ret != 0)
        {
          _err("antenna did not switch to IDLE: %d\n", ret);
        }
      ret = si4463_changestate(priv, NEWSTATE_READY);
      if (ret != 0)
        {
          _err("rx disable failed: %d\n", ret);
        }
      return ret;
    }

  _info("called, buf=%p len=%d chan=%d\n", dev->rxbuf, dev->rxlen, dev->channel);

  /* Setup field 2 length to the MAX receivable length */
  ret = si4463_setprop_short(priv, SI4463_PROP_PKT_FIELD_2_LENGTH_0, 0x1FFF);
  if(ret != 0)
    {
      _err("failed to set field 2 to max length\n");
      return ret;
    }

  /* Go to RX state */
  cmd_rx.channel     = dev->channel;
  cmd_rx.condition   = STARTRX_CONDITION_IMMEDIATE;
  cmd_rx.rx_len      = 0;

  cmd_rx.next_state1 = STARTRX_STATE_RX;    /* State after RX timeout */
  cmd_rx.next_state2 = STARTRX_STATE_READY; /* State after RX valid */
  cmd_rx.next_state3 = STARTRX_STATE_READY; /* State after RX invalid */

  /* Prepare reception state: no bytes received yet (in public and private states)*/
  dev ->rxdone       = 0;
  priv->hdr_done     = false;

  ret = si4463_antennaswitch(priv, ANTENNA_RX);
  if (ret != 0)
    {
      _err("antenna did not switch to RX: %d\n", ret);
      return ret;
    }

  /* Go to RX mode */

  ret = si4463_command(priv, SI4463_CMD_START_RX,
                       &cmd_rx, sizeof(cmd_rx), NULL, 0);
  if(ret == 0)
    {
      ret = si4463_getchipstatus(priv);
    }
  if (ret != 0)
    {
      _err("start rx failed: %d\n", ret);
    }

  return ret;
}

/****************************************************************************
 * Name: si4463_irqworker
 *
 * Description:
 *  IRQ task to avoid executing RTOS code in IRQ context
 *
 ****************************************************************************/
static void si4463_irqworker(FAR void *arg)
{
  struct si4463_dev_s *priv = (struct si4463_dev_s*)arg;
  uint8_t ph,mo,ch;
  volatile int ret;

  /* Get IRQ status */

  ret = si4463_getintstatus(priv, &ph, &mo, &ch);
  if(ret != 0)
    {
      _err("CRITICAL: Could not read interrupt flags\n");
      goto done;
    }
  //_info("!%02X %02X %02X\n", ph,mo,ch);

  /* Dispatch */

  if( (ph & INTSTATUS_PH_PACKET_SENT) == INTSTATUS_PH_PACKET_SENT)
    {
      int val;
      _info("!txc\n");
      ret = si4463_antennaswitch(priv, ANTENNA_IDLE);
      if(ret != 0)
        {
          _err("antenna did not switch to IDLE: %d\n", ret);
        }
      priv->genradio.txstatus = 0;
      ret = nxsem_getvalue(&priv->genradio.txsem, &val);
      if (ret == OK && val <= 0)
        {
          nxsem_post(&priv->genradio.txsem);
        }
    }
  
  if( (ph & INTSTATUS_PH_RX_FIFO_ALMOST_FULL) == INTSTATUS_PH_RX_FIFO_ALMOST_FULL)
    {
      /* RX data is available, take it from fifo and store in target buf*/
      uint32_t room = priv->genradio.rxlen - priv->genradio.rxdone; /* This is how many bytes can be received */
      uint8_t  rxlen;
      bool     flush = false;
      uint32_t rxexpected;

      _info("!rx\n");

      /* Get how many bytes are in the fifo right now */
      ret = si4463_getfifoinfo(priv, false, false, NULL, &rxlen);

      /* If enough data has been received, read the header to find the length of the variable field */
      if(!priv->hdr_done)
        {
          /* The variable packet length is not konwn yet.
           * If we have received enough data, read it.
           * TODO: The packet length should not be returned to the user but received live
           */
#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
          if(rxlen < 1)
            {
              return; //Not enough data received (unlikely)
            }
          si4463_readrxfifo(priv, priv->header, sizeof(priv->header));
          rxlen -= 1;

          rxexpected  = priv->header[0];
          rxexpected += 1; /* this is the length of the header itself */
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2B )
          if(rxlen < 2)
            {
              return; //Not enough data received (unlikely)
            }
          si4463_readrxfifo(priv, priv->header, sizeof(priv->header));
          rxlen -= 2;

          rxexpected   = priv->header[0];
          rxexpected <<= 8;
          rxexpected  |= priv->header[1];
          rxexpected  += 2; /* this is the length of the header itself */
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2L )
          if(rxlen < 2)
            {
              return; //Not enough data received (unlikely)
            }
          si4463_readrxfifo(priv, priv->header, sizeof(priv->header));
          rxlen -= 2;

          rxexpected   = priv->header[1];
          rxexpected <<= 8;
          rxexpected  |= priv->header[2];
          rxexpected  += 2; /* this is the length of the header itself */
#endif
          _info("total on-air packet length: %d\n", rxexpected);
          priv->hdr_done = true;
        }

      /* Now the remaining data is the payload. Trucate it storage-wise, but we
       * will still receive it in full (and discard data that does not fit in
       * the user buffer )
       */
      if(rxlen > room)
        {
          rxlen = room;
          flush = true;
        }

      si4463_readrxfifo(priv, priv->genradio.rxbuf + priv->genradio.rxdone, rxlen);
      priv->genradio.rxdone += rxlen;

      if(flush)
        {
          /* Flush the RX fifo since this data cannot be stored in the user buffer */
          si4463_getfifoinfo(priv, false, true, NULL, NULL);
        }
    }

  if( (ph & INTSTATUS_PH_PACKET_RX) == INTSTATUS_PH_PACKET_RX)
    {
      int val;
      /* Packet reception is complete. TODO read RSSI */
      _info("!rxc\n");
      priv->genradio.rxlen = priv->genradio.rxdone;

      /* Signal the RX waiter semaphore ONLY if it's locked */
      ret = nxsem_getvalue(&priv->genradio.rxsem, &val);
      if (ret == OK && val <= 0)
        {
          nxsem_post(&priv->genradio.rxsem);
        }
      ret = si4463_changestate(priv, NEWSTATE_READY);
    }

  if( (ph & INTSTATUS_PH_CRC_ERROR) == INTSTATUS_PH_CRC_ERROR)
    {
      int val;
      _info("!crc\n");
      priv->genradio.rxlen = -EIO;

      /* Signal the RX waiter semaphore ONLY if it's locked */
      ret = nxsem_getvalue(&priv->genradio.rxsem, &val);
      if (ret == OK && val <= 0)
        {
          nxsem_post(&priv->genradio.rxsem);
        }
      ret = si4463_changestate(priv, NEWSTATE_READY);
    }

//modem ints
  if( (mo & INTSTATUS_MODEM_SYNC_DETECT) == INTSTATUS_MODEM_SYNC_DETECT)
    {
      _info("!sync\n");
    }
  if( (mo & INTSTATUS_MODEM_PREAMBLE_DETECT) == INTSTATUS_MODEM_PREAMBLE_DETECT)
    {
      _info("!pre\n");
    }
#if 0
  if( (mo & INTSTATUS_MODEM_INVALID_PREAMBLE) == INTSTATUS_MODEM_INVALID_PREAMBLE)
    {
      _info("!badpre\n");
    }
#endif
  if( (mo & INTSTATUS_MODEM_RSSI) == INTSTATUS_MODEM_RSSI)
    {
      _info("!rssi\n");
    }
  if( (mo & INTSTATUS_MODEM_RSSI_JUMP) == INTSTATUS_MODEM_RSSI_JUMP)
    {
      _info("!rssijmp\n");
    }
  if( (mo & INTSTATUS_MODEM_INVALID_SYNC) == INTSTATUS_MODEM_INVALID_SYNC)
    {
      _info("!badsync\n");
    }
  if( (mo & INTSTATUS_MODEM_POSTAMBLE_DETECT) == INTSTATUS_MODEM_POSTAMBLE_DETECT)
    {
      _info("!post\n");
    }

//chip ints
  if( (ch & INTSTATUS_CHIP_CAL) == INTSTATUS_CHIP_CAL)
    {
      _info("!cal\n");
    }
  if( (ch & INTSTATUS_CHIP_FIFO_UNDEROVERFLOW) == INTSTATUS_CHIP_FIFO_UNDEROVERFLOW)
    {
      _info("!fiflow\n");
    }
#if 0
  if( (ch & INTSTATUS_CHIP_STATE_CHANGE) == INTSTATUS_CHIP_STATE_CHANGE)
    {
      _info("!state change\n");
    }
#endif
  if( (ch & INTSTTAUS_CHIP_CMD_ERROR) == INTSTTAUS_CHIP_CMD_ERROR)
    {
      _info("!cmderr\n");
    }
  if( (ch & INTSTATUS_CHIP_READY) == INTSTATUS_CHIP_READY)
    {
      _info("!rdy\n");
    }

done:
  /* Re-enable */
  priv->lower->enable(priv->lower, TRUE);
}

/****************************************************************************
 * Name: si4463_interrupt
 *
 * Description:
 *  IRQ handler, called in irq context from the board specific code.
 *
 ****************************************************************************/

static int si4463_interrupt(int irq, FAR void *context, void *arg)
{
  struct si4463_dev_s *priv = (struct si4463_dev_s*)arg;

  DEBUGASSERT(work_available(&priv->irqwork));

  /* Notice that further GPIO interrupts are disabled until the work is
   * actually performed.  This is to prevent overrun of the worker thread.
   * Interrupts are re-enabled in irqworker() when the work is completed.
   */

  priv->lower->enable(priv->lower, FALSE);
  return work_queue(HPWORK, &priv->irqwork, si4463_irqworker,
                    (FAR void *)priv, 0);
}

/****************************************************************************
 * Name: si4463_transmit
 *
 * Description: Transmit a packet. In CW mode a special behaviour is enabled.
 * The written packets must have a length of one byte and ths single byte is
 * taken as a boolean value to enable the CW mode.
 *
 ****************************************************************************/


static void si4463_transmit(FAR struct genradio_dev_s *dev,
                            FAR const uint8_t *packet, size_t len)
{
  struct si4463_cmd_starttx_b1b_s cmd_tx;
  int                             ret;
  FAR struct si4463_dev_s        *priv = (FAR struct si4463_dev_s *)dev;
  size_t                          savelen = len;
  bool                            initial = true;
  uint8_t                         txroom;
  bool                            len_sent;

#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
  if (len > 256)
#else
  if (len > 8192)
#endif
    {
      ret = -E2BIG;
      goto done;
    }

  /* Special behaviour in CW mode */

  if (dev->modulation == RADIO_MOD_CW)
    {
      /* only writes of length one accepted */
      if (len != 1)
        {
          ret = -EINVAL;
          goto done;
        }

      if (packet[0])
        {
          /* Non-zero value: true -> start transmission */
          _info("Starting CW operation\n");
          ret = si4463_command(priv, SI4463_CMD_START_TX,
                               &cmd_tx, sizeof(cmd_tx), NULL, 0);
        }
      else
        {
          /* Zero value: false -> stop transmission */
          _info("Stopping CW operation\n");
          ret = si4463_changestate(priv, NEWSTATE_READY);
        }
      /* Zero means command success, turn this into a successful write
       * of one byte
       */
      if (ret == 0)
        {
          ret = 1;
        }
      goto done;
    }

  /* Special CW mode is now managed, now manage normal packet transmission. */

  /* Process is:
   *   While packet is not complete:
   *     Get room in tx fifo
   *     Copy that many bytes in the FIFO
   *     If this is the first write, trigger TX
   *   This function returns as soon as all bytes have been written in the FIFO.
   *   Completion notification is done via an interrupt.
   */

  /* First of all go to the READY state */
  //_info("tx start: channel %d\n", dev->channel);
  ret = si4463_changestate(priv, NEWSTATE_READY);
  if (ret != 0)
    {
      _err("change state failed: %d\n", ret);
      goto done;
    }

  len_sent = false;
  while(len > 0)
    {
      ret = si4463_getfifoinfo(priv, initial, false, &txroom, NULL);

      if (ret != 0)
        {
          _err("get fifo info failed: %d\n", ret);
          goto done;
        }
      //_info("fifo room: %d\n", txroom);

      /* Send the header - length byte or bytes, then define the length of
       * the second field. */
      if(!len_sent)
        {
          uint8_t lenbuf[2];
          int     lenlen = 0;
#if defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L1 )
          lenbuf[0] = len & 0xFF;
          lenlen = 1;
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2B )
          lenbuf[0] = len >> 8;
          lenbuf[1] = len & 0xFF;
          lenlen = 2;
#elif defined( CONFIG_GENERICRADIO_SI4463_FRAMING_L2L )
          lenbuf[0] = len & 0xFF;
          lenbuf[1] = len >> 8;
          lenlen = 2;
#endif
          ret = si4463_setprop_short(priv, 
                                     SI4463_PROP_PKT_FIELD_2_LENGTH_0, len);
          if(ret != 0)
            {
              _err("start tx failed: %d\n", ret);
              goto done;
            }

          if(txroom < lenlen)
            {
              _err("no room to write header in fifo\n");
              ret = -EIO;
              goto done;
            }
          si4463_writetxfifo(priv, lenbuf, lenlen);
          txroom -= lenlen;
          len_sent = true;
        }

      /* Decide how many bytes to write */
      if(len < txroom)
        {
          txroom = len;
        }

      ret = si4463_writetxfifo(priv, packet, txroom);
      if(ret < 0)
        {
          goto done;
        }

      if(initial)
        {
          ret = si4463_antennaswitch(priv, ANTENNA_TX);
          if(ret != 0)
            {
              _err("antenna did not switch to TX: %d\n", ret);
              goto done;
            }
          /* Start transmission right after the first fifo fill */
          initial = false;
          cmd_tx.channel   = dev->channel;
          cmd_tx.condition = STARTTX_COND_TXCOMPLETE_READY | STARTTX_COND_START_IMMEDIATE;
          cmd_tx.txlen = 0; //htobes(len);
          ret = si4463_command(priv, SI4463_CMD_START_TX, &cmd_tx, sizeof(cmd_tx), NULL, 0);
          if(ret == 0)
            {
              ret = si4463_getchipstatus(priv);
            }
          if(ret != 0)
            {
              _err("start tx failed: %d\n", ret);
              goto done;
            }
          //_info("Start TX success\n");
        }

      len    -= txroom;
      packet += txroom;
    }

  //_info("done\n");

  ret = savelen;

done:
  priv->genradio.txstatus = ret;
  return;
}

/****************************************************************************
 * Name: si4463_ioctl
 *
 * Description:
 *
 ****************************************************************************/

static int si4463_ioctl(FAR struct genradio_dev_s *dev, int cmd, unsigned long arg)
{
  int ret = -ENOTTY;
  FAR struct si4463_dev_s *priv = (FAR struct si4463_dev_s *)dev;

  return ret;
}

/****************************************************************************
 * Public functions
 ****************************************************************************/

FAR struct genradio_dev_s *si4463_init(FAR struct spi_dev_s *spi, int devid,
                                       uint32_t xtal, int papin, int lnapin,
                                       bool activelow,
                                       FAR const struct si4463_lower_s *lower)
{
  FAR struct si4463_dev_s *dev;
  int ret;
  uint32_t what = 0;

  _info("Start, xtal=%lu\n", xtal);

  dev = kmm_zalloc(sizeof(struct si4463_dev_s));

  if (!dev)
    {
      return NULL;
    }

  /* Attach irq */

  if (lower->attach(lower, si4463_interrupt, dev) != OK)
    {
      goto retfree;
    }

  dev->genradio.ops = &si4463_devops;

  /* Initialize semaphores */

  nxsem_init(&dev->genradio.rxsem, 0, 0);
  nxsem_init(&dev->genradio.txsem, 0, 0);

  /* These semaphores are all used for signaling and, hence, should
   * not have priority inheritance enabled.
   */

  nxsem_setprotocol(&dev->genradio.rxsem, SEM_PRIO_NONE);
  nxsem_setprotocol(&dev->genradio.txsem, SEM_PRIO_NONE);

  dev->lower    = lower;
  dev->spi      = spi;
  dev->spiid    = devid;
  dev->xtal     = xtal;

  ret = si4463_initialize(dev, papin, lnapin, activelow);
  if(ret != 0)
    {
      lower->attach(lower, NULL, NULL);
      goto retfree;
    }

#ifndef CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG
  dev->requested_modtype = RADIO_MOD_2FSK;
  what |= SI4463_APPLY_MODTYPE;

  /* Default device params */
#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_FREQ
  dev->requested_basefreq = CONFIG_GENERICRADIO_SI4463_DEFAULT_FREQ;
  what |= SI4463_APPLY_BASEFREQ;
#endif

#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_RATE
  dev->requested_datarate = CONFIG_GENERICRADIO_SI4463_DEFAULT_RATE * 1000;
  what |= SI4463_APPLY_DATARATE;
#endif

#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_DEV
  dev->requested_deviation = CONFIG_GENERICRADIO_SI4463_DEFAULT_DEV * 1000;
  what |= SI4463_APPLY_DEVIATION;
#endif

#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_CHSPC
  dev->requested_chanspacing = CONFIG_GENERICRADIO_SI4463_DEFAULT_CHSPC;
  what |= SI4463_APPLY_CHANSPACING;
#endif

  ret = si4463_applysettings(dev, what);
  if(ret < 0)
    {
      _err("Failed to apply RF settings\n");
      goto retfree;
    }

  si4463_setpreamblelength(&dev->genradio, CONFIG_GENERICRADIO_SI4463_DEFAULT_PREAMBLELEN);

#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_SYNCWORD_A55A
  si4463_setsyncword(&dev->genradio, (uint8_t*)"\xA5\x5A", 16);
#endif
#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_SYNCWORD_4242
  si4463_setsyncword(&dev->genradio, (uint8_t*)"\x42\x42", 16);
#endif
#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_SYNCWORD_A55AA55A
  si4463_setsyncword(&dev->genradio, (uint8_t*)"\xA5\x5A\xA5\x5A", 32);
#endif
#ifdef CONFIG_GENERICRADIO_SI4463_DEFAULT_SYNCWORD_42424242
  si4463_setsyncword(&dev->genradio, (uint8_t*)"\x42\x42\x42\x42", 32);
#endif

  si4463_settxpower(&dev->genradio, 0x7F);

#endif //CONFIG_GENERICRADIO_SI4463_USE_WDS_CONFIG

  /* TODO */
  
  /* Enable the IRQ */
  
  dev->lower->enable(dev->lower, TRUE);

  return &dev->genradio;

retfree:
  free(dev);
  return NULL;
}

