/****************************************************************************
 * drivers/wireless/generic/upperhalf.c
 *
 *   Copyright (C) 2017 Sebastien Lorquet. All rights reserved.
 *   Author: Sebastien Lorquet <sebastien@lorquet.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name NuttX nor the names of its contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/

/****************************************************************************
 * Included Files
 ****************************************************************************/

#include <nuttx/config.h>
#include <debug.h>
#include <assert.h>
#include <stdint.h>
#include <errno.h>
#include <nuttx/semaphore.h>

#include <nuttx/config.h>

#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <nuttx/arch.h>
#include <nuttx/kmalloc.h>
#include <nuttx/wireless/generic/ioctl.h>
#include <nuttx/wireless/generic/genradio.h>

/****************************************************************************
 * Pre-Processor Definitions
 ****************************************************************************/
 
/****************************************************************************
 * Private Types
 ****************************************************************************/

struct genradio_devwrapper_s
{
  FAR struct genradio_dev_s *child;
  sem_t devsem;                         /* Device access serialization semaphore */
  bool opened;                          /* This device can only be opened once */
};

/****************************************************************************
 * Private Function Prototypes
 ****************************************************************************/

static void genradio_semtake(FAR struct genradio_devwrapper_s *dev);
static int  genradio_open(FAR struct file *filep);
static int  genradio_close(FAR struct file *filep);
static ssize_t genradio_read(FAR struct file *filep, FAR char *buffer,
              size_t len);
static ssize_t genradio_write(FAR struct file *filep,
              FAR const char *buffer, size_t len);
static int  genradio_ioctl(FAR struct file *filep, int cmd,
              unsigned long arg);

/****************************************************************************
 * Private Data
 ****************************************************************************/

static const struct file_operations genradio_fops =
{
  genradio_open , /* open */
  genradio_close, /* close */
  genradio_read , /* read */
  genradio_write, /* write */
  NULL,                 /* seek */
  genradio_ioctl  /* ioctl */
#ifndef CONFIG_DISABLE_POLL
  , NULL               /* poll */
#endif
#ifndef CONFIG_DISABLE_PSEUDOFS_OPERATIONS
  , NULL               /* unlink */
#endif
};

/****************************************************************************
 * Private Functions
 ****************************************************************************/

/****************************************************************************
 * Name: genradio_semtake
 *
 * Description:
 *   Acquire the semaphore used for access serialization.
 *
 ****************************************************************************/

static void genradio_semtake(FAR struct genradio_devwrapper_s *dev)
{
  int ret;
  /* Take the semaphore (perhaps waiting) */

  while ( (ret = nxsem_wait(&dev->devsem)) != 0)
    {
      /* The only case that an error should occur here is if
       * the wait was awakened by a signal.
       */

      DEBUGASSERT(ret == -EINTR);
    }
}

/****************************************************************************
 * Name: genradio_semgive
 *
 * Description:
 *   Release the semaphore used for access serialization.
 *
 ****************************************************************************/

static inline void genradio_semgive(FAR struct genradio_devwrapper_s *dev)
{
  nxsem_post(&dev->devsem);
}

/****************************************************************************
 * Name: genradio_open
 *
 * Description:
 *   Open the generic radio character device.
 *
 ****************************************************************************/

static int genradio_open(FAR struct file *filep)
{
  FAR struct inode *inode;
  FAR struct genradio_devwrapper_s *dev;

  DEBUGASSERT(filep != NULL && filep->f_inode != NULL);
  inode = filep->f_inode;
  dev   = inode->i_private;
  DEBUGASSERT(dev != NULL);

  /* Get exclusive access to the driver data structures */

  genradio_semtake(dev);

  if (dev->opened)
    {
      /* Already opened */

      return -EMFILE;
    }
  else
    {
      /* Enable interrupts (only rx for now)*/

      //dev->lower->enable(dev->lower, TRUE);

      dev->opened = true;
    }

  genradio_semgive(dev);
  return OK;
}

/****************************************************************************
 * Name: genradio_close
 *
 * Description:
 *   Close the generic radio character device.
 *
 ****************************************************************************/

static int genradio_close(FAR struct file *filep)
{
  FAR struct inode *inode;
  FAR struct genradio_devwrapper_s *dev;
  int ret = OK;

  DEBUGASSERT(filep != NULL && filep->f_inode != NULL);
  inode = filep->f_inode;
  dev   = inode->i_private;
  DEBUGASSERT(dev != NULL);

  /* Get exclusive access to the driver data structures */

  genradio_semtake(dev);

  if (!dev->opened)
    {
      /* Driver has not been opened */

      ret = -EIO;
    }
  else
    {
      /* Disable interrupts */

      //dev->lower->enable(dev->lower, FALSE);

      dev->opened = false;
    }

  genradio_semgive(dev);
  return ret;
}

/****************************************************************************
 * Name: genradio_read
 *
 * Description:
 *   Return the last received packet.
 *   TODO: Return a packet from the receive queue.
 *
 ****************************************************************************/

static ssize_t genradio_read(FAR struct file *filep, FAR char *buffer,
                             size_t len)
{
  FAR struct inode *inode;
  FAR struct genradio_devwrapper_s *dev;
  FAR struct timespec timeout;
  int ret = OK;

  DEBUGASSERT(filep != NULL && filep->f_inode != NULL);
  inode = filep->f_inode;
  dev   = inode->i_private;
  DEBUGASSERT(dev != NULL && buffer != NULL);

  /* Setup receive timeout */
  clock_gettime(CLOCK_REALTIME, &timeout);

  timeout.tv_sec += 5;
  timeout.tv_nsec = 0;

  /* Tell the device where to store the next received packet */

  dev->child->rxbuf = buffer;
  dev->child->rxlen = len;

  /* Trigger reception */

  ret = dev->child->ops->rxenable(dev->child, 1);

  if (ret < 0)
    {
      return ret;
    }
  
  /* if no packet is received, this will produce -EAGAIN
   * The user is responsible for trying again until sth arrives
   */

  ret = nxsem_timedwait(&dev->child->rxsem, &timeout);
  if(ret != 0)
    {
      wlerr("Radio Device timedout on Rx\n");
      //TODO cancel transmission and reinit semaphore
      nxsem_init(&dev->child->rxsem, false, 0);
      goto done;
    }
  //ret = nxsem_wait(&dev->child->rxsem);


  ret = dev->child->rxlen;

done:
  //disable further reads until called again.
  dev->child->ops->rxenable(dev->child, 0);
  return ret;
}

/****************************************************************************
 * Name: genradio_write
 *
 * Description:
 *   Send a packet immediately.
 *   TODO: Put a packet in the send queue. The packet will be sent as soon
 *   as possible.
 *
 ****************************************************************************/

static ssize_t genradio_write(FAR struct file *filep,
                                   FAR const char *buffer, size_t len)
{
  FAR struct inode *inode;
  FAR struct genradio_devwrapper_s *dev;
  FAR struct timespec timeout;
  int ret = OK;

  DEBUGASSERT(filep != NULL && filep->f_inode != NULL);
  inode = filep->f_inode;
  dev   = inode->i_private;
  DEBUGASSERT(dev != NULL);

  /* TODO: Make this an option or have a new method of doing timeout from
   * ioctrl.
   */
  clock_gettime(CLOCK_REALTIME, &timeout);
  timeout.tv_sec += 5;
  timeout.tv_nsec = 0;

  dev->child->txstatus = 0;

  /* Trigger transmission */

  dev->child->ops->transmit(dev->child, (const uint8_t*)buffer, len);

  /* Wait for completion of the transmission */

  ret = nxsem_timedwait(&dev->child->txsem, &timeout);

  if(ret != 0)
    {
      wlerr("Radio Device timedout on Tx, errno=%d\n",errno);
      return ret;
    }

  ret = dev->child->txstatus;
  if(ret == 0)
    {
      ret = len;
    }

  /* Okay, tx completion interrupt received.
   * Check transmission status to decide success. */

  return ret;
}

/****************************************************************************
 * Name: genradio_ioctl
 *
 * Description:
 *   Control the generic radio device.
     This is where the real operations happen.
 *
 ****************************************************************************/

static int genradio_ioctl(FAR struct file *filep, int cmd,
                             unsigned long arg)
{
  FAR struct inode *inode;
  FAR struct genradio_devwrapper_s *dev;
  FAR struct genradio_dev_s *child;
  FAR struct genradio_synciocparam_s *syncarg;

  int ret = OK;

  DEBUGASSERT(filep != NULL && filep->f_inode != NULL);
  inode = filep->f_inode;
  dev   = inode->i_private;
  DEBUGASSERT(dev != NULL && dev->child != NULL);
  child = dev->child;

  switch (cmd)
    {

      case GRIOCSETBFREQ:
        ret = child->ops->setbasefreq(child, (uint32_t)arg);
        break;

      case GRIOCSETCHSPC:
        ret = child->ops->setchanspacing(child, (uint32_t)arg);
        break;

      case GRIOCSETCHAN:
        ret = child->ops->setchannel(child, (int32_t)arg);
        break;

      case GRIOCSETDRATE:
        ret = child->ops->setdatarate(child, (uint32_t)arg);
        break;

      case GRIOCSETMOD:
        ret = child->ops->setmodulation(child, (uint32_t)arg);
        break;

      case GRIOCSETMODI:
        ret = child->ops->setmodulationindex(child, (uint32_t)arg);
        break;

      case GRIOCSETDEV:
        ret = child->ops->setdeviation(child, (uint32_t)arg);
        break;

      case GRIOCSETGBT:
        ret = child->ops->setfilterbt(child, (uint32_t)arg);
        break;

      case GRIOCSETPLEN:
        ret = child->ops->setpreamblelength(child, (uint32_t)arg);
        break;

      case GRIOCSETSYNC:
        syncarg = (FAR struct genradio_synciocparam_s*)arg;
        ret = child->ops->setsyncword(child, syncarg->data, syncarg->len);
        break;

      case GRIOCSETRXBW:
        ret = child->ops->setrxbandwidth(child, (uint32_t)arg);
        break;

      case GRIOCGETRSSI:
        ret = child->ops->getrssi(child, (FAR int32_t*)arg);
        break;

      case GRIOCGETLQI:
        ret = child->ops->getlqi(child, (FAR int32_t*)arg);
        break;

      case GRIOCSETTXPW:
        ret = child->ops->settxpower(child, (int32_t)arg);
        break;

/* Trivial getters. Device functions only called if they exist, otherwise
 * a generic version is used.
 */

      case GRIOCGETBFREQ:
        if (child->ops->getbasefreq)
          {
            ret = child->ops->getbasefreq(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->basefreq;
          }
        break;

      case GRIOCGETCHSPC:
        if (child->ops->getchanspacing)
          {
            ret = child->ops->getchanspacing(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->chanspacing;
          }
        break;

      case GRIOCGETCHAN:
        if (child->ops->getchannel)
          {
            ret = child->ops->getchannel(child, (FAR int32_t*)arg);
          }
        else
          {
            *((FAR int32_t*)arg) = child->channel;
          }
        break;

      case GRIOCGETDRATE:
        if (child->ops->getdatarate)
          {
            ret = child->ops->getdatarate(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->datarate;
          }
        break;

      case GRIOCGETMOD:
        if (child->ops->getmodulation)
          {
            ret = child->ops->getmodulation(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->modulation;
          }
        break;

      case GRIOCGETMODI:
        if (child->ops->getmodulationindex)
          {
            ret = child->ops->getmodulationindex(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->modindex;
          }
        break;

      case GRIOCGETDEV:
        if (child->ops->getdeviation)
          {
            ret = child->ops->getdeviation(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->deviation;
          }
        break;

      case GRIOCGETGBT:
        if (child->ops->getfilterbt)
          {
            ret = child->ops->getfilterbt(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->filterbt;
          }
        break;

      case GRIOCGETPLEN:
        if (child->ops->getpreamblelength)
          {
            ret = child->ops->getpreamblelength(child, (uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->preamblelen;
          }
        break;

      case GRIOCGETRXBW:
        if (child->ops->getrxbandwidth)
          {
            ret = child->ops->getrxbandwidth(child, (FAR uint32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->rxbandwidth;
          }
        break;

      case GRIOCGETTXPW:
        if (child->ops->gettxpower)
          {
            ret = child->ops->gettxpower(child, (FAR int32_t*)arg);
          }
        else
          {
            *((FAR uint32_t*)arg) = child->txpower;
          }
        break;

      case GRIOCGETSYNC:
        syncarg = (FAR struct genradio_synciocparam_s*)arg;
        if (child->ops->getsyncword)
          {
            ret = child->ops->getsyncword(child, syncarg->data, &syncarg->len);
          }
        else
          {
            if(syncarg->len < child->syncwordlen)
              {
                _err("Not enough room\n");
                ret = -EINVAL;
              }
            else if(syncarg->len > GENRADIO_SYNCWORD_MAX)
              {
                /* sanity check, might detect uninitialized values */
                _err("ioctl arg probably not initialized\n");
                ret = -EINVAL;
              }
            else
              {
                memcpy(syncarg->data, child->syncword, BYTESFORBITS(child->syncwordlen));
                syncarg->len = child->syncwordlen;
              }
          }
        break;

      default:
        if(child->ops->ioctl)
          {
            ret = child->ops->ioctl(child, cmd, arg);
          }
        else
          {
            ret = -ENOTTY;
          }
        break;
    }

  return ret;
}

/****************************************************************************
 * Public functions
 ****************************************************************************/

/****************************************************************************
 * Name: genradio_register
 *
 * Description:
 *   Register a character driver to access the generic radio from
 *   user-space
 *
 * Input Parameters:
 *   radio - Pointer to the radio struct to be registerd.
 *   devname - The name of the generic radio to be registered.
 *
 * Returned Values:
 *   Zero (OK) is returned on success.  Otherwise a negated errno value is
 *   returned to indicate the nature of the failure.
 *
 ****************************************************************************/
int genradio_register(FAR struct genradio_dev_s *radio, FAR const char *devname)
{
  FAR struct genradio_devwrapper_s *dev;

  dev = kmm_zalloc(sizeof(struct genradio_devwrapper_s));
  if (dev == NULL)
    {
      return -ENOMEM;
    }

  dev->child = radio;

  nxsem_init(&dev->devsem, 0, 1); /* Allow the device to be opened once before blocking */

  return register_driver(devname, &genradio_fops, 0666, dev);
}


