/****************************************************************************
 * include/nuttx/wireless/generic/genradio.h
 *
 *   Copyright (C) 2017 Sebastien Lorquet. All rights reserved.
 *   Author: Sebastien Lorquet <sebastien@lorquet.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name NuttX nor the names of its contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/

#ifndef __INCLUDE_NUTTX_WIRELESS_GENERIC_GENRADIO_H
#define __INCLUDE_NUTTX_WIRELESS_GENERIC_GENRADIO_H

/****************************************************************************
 * Included Files
 ****************************************************************************/

#include <nuttx/config.h>
#include <stdint.h>
#include <semaphore.h>

/****************************************************************************
 * Pre-Processor Definitions
 ****************************************************************************/

/* Modulation types */

enum
{
  RADIO_MOD_2FSK,  /* Frequency shift keying */
  RADIO_MOD_2GFSK, /* Gaussian Frequency shift keying */
  RADIO_MOD_4FSK,  /* Frequency shift keying, 4 tones */
  RADIO_MOD_4GFSK, /* Gaussian Frequency shift keying, 4 tones */
  RADIO_MOD_CW,    /* Continuous wave, for testing */
  RADIO_MOD_FSK_LAST = RADIO_MOD_CW,
  RADIO_MOD_ASK,   /* Amplitude shift keying */
  RADIO_MOD_OOK,   /* On-Off keying */
  RADIO_MOD_MSK,   /* Minimum Shift Keying */
  RADIO_MOD_GMSK,  /* Gaussian Minimum Shift Keying */
  RADIO_MOD_BPSK,  /* Binary Phase Shift Keying */
  RADIO_MOD_QPSK,  /* Quadrature Phase Shift Keying */
  RADIO_MOD_LORA,  /* Proprietary chirped LoRa modulation */
  RADIO_MOD_COUNT
};

/* Maximum number of bits in the packet sync word */
#define GENRADIO_SYNCWORD_MAX 64

/* Return the number of bytes required to store this count of bits */
#define BYTESFORBITS(bits) ((bits)?(1+(bits-1)>>3):0)

#define GENRADIO_SYNCWORD_MAXBYTES BYTESFORBITS(GENRADIO_SYNCWORD_MAX)

/****************************************************************************
 * Public Types
 ****************************************************************************/

struct genradio_dev_s;

struct genradio_devops_s
{
  /* ======= */
  /* General */
  /* ======= */

  /* Get or set the base transmission or reception frequency in Hz (chan 0), max 4 294 967 295 Hz (4 GHz) */
  CODE int (*setbasefreq)(FAR struct genradio_dev_s *dev, uint32_t Hz);
  CODE int (*getbasefreq)(FAR struct genradio_dev_s *dev, FAR uint32_t *Hz);

  /* TODO Get or set the main transmission or reception frequency in uHz, max 18 446 744 073 709,551 615 Hz (18 THz) */
  
  /* Get or set the channel spacing in mHz */
  CODE int (*setchanspacing)(FAR struct genradio_dev_s *dev, uint32_t mHz);
  CODE int (*getchanspacing)(FAR struct genradio_dev_s *dev, FAR uint32_t *mHz);

  /* Get or set the channel number (range depends on chip) */
  CODE int (*setchannel)(FAR struct genradio_dev_s *dev, int32_t chan);
  CODE int (*getchannel)(FAR struct genradio_dev_s *dev, FAR int32_t *chan);

  /* Get or set the data rate in thousands of bits/second, max 4 294 967.295 bit/s */
  CODE int (*setdatarate)(FAR struct genradio_dev_s *dev, uint32_t mbps);
  CODE int (*getdatarate)(FAR struct genradio_dev_s *dev, FAR uint32_t *mbps);
  
  /* Send custom commands to a module */
  CODE int (*ioctl)(FAR struct genradio_dev_s *dev, int cmd, unsigned long arg);
  
  /* ========== */
  /* Modulation */
  /* ========== */
  
  /* Get or Set the modulation used by the module */
  CODE int (*setmodulation)(FAR struct genradio_dev_s *dev, uint32_t mod);
  CODE int (*getmodulation)(FAR struct genradio_dev_s *dev, FAR uint32_t *mod);

  /* For ASK modulation, set the modulation index in ppm, max 4294967296 ppm = 4294.967296 */
  CODE int (*setmodulationindex)(FAR struct genradio_dev_s *dev, uint32_t ppm);
  CODE int (*getmodulationindex)(FAR struct genradio_dev_s *dev, FAR uint32_t *ppm);

  /* For FSK modulations, set the frequency deviation FROM THE CARRIER,
   * in thousands of an hertz, max 4 294 967.295 Hz.
   * For 4-FSK this is the OUTER deviation d, where the other symbols have
   * deviations -d, -d/3, d/3, d.
   */
  CODE int (*setdeviation)(FAR struct genradio_dev_s *dev, uint32_t mHz);
  CODE int (*getdeviation)(FAR struct genradio_dev_s *dev, FAR uint32_t *mHz);

  /* For gaussian filtered modulations, set the filter characteristics
   * Value is between 0 and 1 with usual values 0.5 and 0.1*/
  CODE int (*setfilterbt)(FAR struct genradio_dev_s *dev, uint32_t frac);
  CODE int (*getfilterbt)(FAR struct genradio_dev_s *dev, FAR uint32_t *frac);

  /* ======= */
  /* Framing */
  /* ======= */

  /* Define the length of the preamble field in the packet structure */
  CODE int (*setpreamblelength)(FAR struct genradio_dev_s *dev, uint32_t bits);
  CODE int (*getpreamblelength)(FAR struct genradio_dev_s *dev, uint32_t *bits);

  /* Define the sync word used to recognize the start of valid packets */
  CODE int (*setsyncword)(FAR struct genradio_dev_s *dev, FAR uint8_t *data, uint32_t bits);
  CODE int (*getsyncword)(FAR struct genradio_dev_s *dev, FAR uint8_t *data, uint32_t *bits); /* WARNING bits must be initialized to max buffer size*/

  /* ========= */
  /* Reception */
  /* ========= */

  /* Get or Set the receiver bandwidth in Hz/1000, max 4 294 967.295*/
  CODE int (*setrxbandwidth)(FAR struct genradio_dev_s *dev, uint32_t mHz);
  CODE int (*getrxbandwidth)(FAR struct genradio_dev_s *dev, FAR uint32_t *mHz);
  
  /* Get the last calculated RSSI in mB (hundredths of a dB) */
  CODE int (*getrssi)(FAR struct genradio_dev_s *dev, FAR int32_t *mB);
  
  /* Get the last calculated LQI in arbitrary units */
  CODE int (*getlqi)(FAR struct genradio_dev_s *dev, FAR int32_t *lqi);

  /* Enable reception of packets */
  CODE int (*rxenable)(FAR struct genradio_dev_s *dev, int state);

  /* ============ */
  /* Transmission */
  /* ============ */

  /* Get or Set the transmitted signal power in mB, hundredths of a dB) */
  CODE int (*settxpower)(FAR struct genradio_dev_s *dev, int32_t mB);
  CODE int (*gettxpower)(FAR struct genradio_dev_s *dev, FAR int32_t *mB);

  /* Trigger transmission of a packet over the air */
  CODE void (*transmit)(FAR struct genradio_dev_s *dev, FAR const uint8_t *packet, size_t len);

};

struct genradio_dev_s
{
  FAR const struct genradio_devops_s *ops;

  /* Common radio parameters accessed by the above API */

  uint32_t basefreq;   /* Hz */
  uint32_t chanspacing; /* mHz */
  int32_t  channel;
  uint32_t datarate;    /* bps/1000 */
  uint32_t modindex;    /* ppm */
  uint32_t deviation;   /* Hz/1000 DEVIATION FROM THE CARRIER*/
  uint32_t filterbt;    /* 16.16 fixed point */
  uint32_t preamblelen; /* bits */
  uint8_t  syncword[GENRADIO_SYNCWORD_MAX];
  uint32_t syncwordlen; /* bits */
  uint16_t modulation;  /* enum */
  uint32_t rxbandwidth; /* Hz/1000 */
  int32_t  txpower;     /* mBm = dBm/100 */

  /* Packet reception management */

  FAR uint8_t *rxbuf;  /* Buffer in which to store the next received packet */
  uint32_t     rxlen;  /* Length of received data */
  uint32_t     rxdone; /* How many bytes have been received up to now */
  sem_t        rxsem;  /* post()ed after a packet is received, init by lower half */

  /* Packet transmission management */

  sem_t        txsem; /* post()ed when packet transmission is complete, init by lower half */
  int          txstatus; /* transmission status, an errno value */
};

#ifdef __cplusplus
#define EXTERN extern "C"
extern "C"
{
#else
#define EXTERN extern
#endif

/****************************************************************************
 * Public Function Prototypes
 ****************************************************************************/

/****************************************************************************
 * Name: genradio_register
 *
 * Description:
 *   Register a character driver to access the generic radio from
 *   user-space
 *
 * Input Parameters:
 *   radio - Pointer to the radio struct to be registerd.
 *   devname - The name of the generic radio to be registered.
 *
 * Returned Values:
 *   Zero (OK) is returned on success.  Otherwise a negated errno value is
 *   returned to indicate the nature of the failure.
 *
 ****************************************************************************/
int genradio_register(FAR struct genradio_dev_s *radio, FAR const char *devname);

#undef EXTERN
#ifdef __cplusplus
}
#endif

#endif /* __INCLUDE_NUTTX_WIRELESS_GENERIC_GENERIC_RADIO_H */
